import numpy as np

from tifffile import imwrite #, tiffcomment, imread


from datetime import datetime
from typing import List, Union
# from aicsimageio import AICSImage
# from aicsimageio.writers import OmeTiffWriter
from pathlib import Path


# from aicssegmentation.util.filesystem import FileSystemUtilities
# from aicssegmentation.exceptions import ArgumentNullError
from infer_subc.utils.filesystem import FileSystemUtilities
from infer_subc.exceptions import ArgumentNullError
from infer_subc.workflow.workflow import Workflow
from infer_subc.workflow.workflow_definition import WorkflowDefinition

# from infer_subc.core.file_io import reader_function
from infer_subc.core.file_io import reader_function

PathLike = Union[str, Path]
SUPPORTED_FILE_EXTENSIONS = ["tiff", "tif", "czi"]


# TODO: fix channel index.
class BatchWorkflow:
    """
    Represents a batch of workflows to process.
    This class provides the functionality to run batches of workflows using multiple image inputs from a input directory
    according to the steps defined in its WorkflowDefinition.
    """

    def __init__(
        self,
        workflow_definitions: List[WorkflowDefinition],
        input_dir: Union[str, Path],
        output_dir: Union[str, Path],
        segmentation_names: List[str],  # JAH: add segmentation name for export
        channel_index: int = -1,  # JAH: change so all negative indices return ALL the channels/zslices
    ):
        if workflow_definitions is None:
            raise ArgumentNullError("workflow_definitions")
        if segmentation_names is None:
            raise ArgumentNullError("segmentation_names")
        if input_dir is None:
            raise ArgumentNullError("input_dir")
        if output_dir is None:
            raise ArgumentNullError("output_dir")

        self._workflow_definitions = workflow_definitions
        self._input_dir = Path(input_dir)
        self._segmentation_names = segmentation_names

        if not self._input_dir.exists():
            raise ValueError("The input directory does not exist")

        self._output_dir = Path(output_dir)
        self._channel_index = channel_index
        self._processed_files: int = 0
        self._failed_files: int = 0
        self._log_path: Path = self._output_dir / f"log_{datetime.now().strftime('%Y%m%d_%H%M%S')}.txt"

        # Create the output directory at output_dir if it does not exist already
        if not self._output_dir.exists():
            FileSystemUtilities.create_directory(self._output_dir)

        self._input_files = self._get_input_files(self._input_dir, SUPPORTED_FILE_EXTENSIONS)
        self._execute_generator = self._execute_generator_func()

    @property
    def total_files(self) -> int:
        return len(self._input_files) * len(self._workflow_definitions)

    @property
    def processed_files(self) -> int:
        return self._processed_files

    @property
    def failed_files(self) -> int:
        return self._failed_files

    @property
    def input_dir(self) -> Path:
        return self._input_dir

    @property
    def output_dir(self) -> Path:
        return self._output_dir

    @property
    def segmentation_names(self) -> List[str]:
        return self._segmentation_names

    def is_done(self) -> bool:
        """
        Indicates whether all files / steps have been executed

        Use this to know when the batch workflow is complete if manually executing the workflow
        with execute_next()

        Returns
            (bool): True if all files/steps have been executed, False if not
        """
        return self._processed_files == self.total_files

    def execute_all(self):
        if self.is_done():
            print("No files left to process")
            return

        print("Starting batch workflow...")
        print(f"Found {self.total_files} files X workflows to process.")

        while not self.is_done():
            self.execute_next()

        self.write_log_file_summary()

        print(f"Batch workflow complete. Check {self._log_path} for output log and summary.")

    def execute_next(self):
        if self.is_done():
            print("No files left to process")
            return

        next(self._execute_generator)

    def _execute_generator_func(self):
        for f in self._input_files:
            msg = f"\nprocessing::: {f.name} :"
            self._write_to_log_file(msg)
            for wf, seg_nm in zip(self._workflow_definitions, self._segmentation_names):
                try:
                    # read and format image in the way we expect
                    # read_image = AICSImage(f)
                    # image_from_path = self._format_image_to_3d(read_image)
                    image_from_path = self._format_image_to_3d(f)

                    # Run workflow on image
                    msg = f": {seg_nm} :"
                    self._write_to_log_file(msg)

                    workflow = Workflow(wf, image_from_path)
                    while not workflow.is_done():
                        workflow.execute_next()
                        result = workflow.get_most_recent_result()
                        # print(f"current result shape + type: {result.shape}, {result.dtype}")
                        yield

                    # Save output
                    # output_path = self._output_dir / f"{f.stem}.segmentation.tiff"
                    output_path = self._output_dir / f"{f.stem}-{seg_nm}.tiff"

                    result = workflow.get_most_recent_result()
                    result = self._format_output(result)

                    ret = imwrite(
                        output_path,
                        result,
                        dtype=result.dtype,
                        # metadata={
                        #     "axes": dimension_order,
                        #     # "physical_pixel_sizes": physical_pixel_sizes,
                        #     # "channel_names": channel_names,
                        # },
                    )
                    # if len(result.shape) == 3:
                    #     # TODO:  replace with. tifffile writer ...
                    #     OmeTiffWriter.save(data=self._format_output(result), uri=output_path, dim_order="ZYX")
                    # else:
                    #     OmeTiffWriter.save(data=self._format_output(result), uri=output_path, dim_order="CZYX")

                    msg = f"SUCCESS: {f}:{seg_nm}. >>>> {output_path.name}"
                    print(msg)
                    self._write_to_log_file(msg)

                except Exception as ex:
                    self._failed_files += 1
                    msg = f"FAILED: {f}:{seg_nm}, ERROR: {ex}"
                    print(msg)
                    self._write_to_log_file(msg)
                finally:
                    self._processed_files += 1

                yield

    def write_log_file_summary(self):
        """
        Write a log file to the output folder.
        """
        if self._processed_files == 0:
            report = (
                f"There were no files to process in the input directory to process \n "
                f"Using the Workflow: {self._workflow_definition.name}"
            )
        else:
            files_processed = self._processed_files - self._failed_files
            wfs = ", ".join([wfd.name for wfd in self._workflow_definitions])
            report = (
                f"{files_processed}/{self._processed_files} files were successfully processed \n "
                f"Using the Workflows: {wfs}"
            )
        self._write_to_log_file(report)

    # def _format_image_to_3d(self, image: AICSImage) -> np.ndarray:
    def _format_image_to_3d(self, image_path: PathLike) -> np.ndarray:
        """
        Format images in the way that aics-segmention expects for most workflows (3d, zyx)

        Params:
            image_path (Path or str): image to format

        Returns
            np.ndarray: segment-able image for aics-segmentation
        """
        # TODO: make this reading more performant
        # if len(image.scenes) > 1:
        #     raise ValueError("Multi-Scene images are unsupported")

        # if image.dims.T > 1:
        #     raise ValueError("Timelapse images are unsupported.")

        # if image.dims.C > 1:
        #     return image.get_image_data("ZYX", C=self._channel_index)

        # return image.get_image_data("ZYX")
        # JAH: refactdor to use reader_function
        data, meta, layer_type = reader_function(image_path)[0]

        if isinstance(image_path, str):
            image_path = Path(image_path)
        name = image_path.stem

        channel_names = meta.pop("name")  # list of names for each layer
        meta["channel_names"] = channel_names
        meta["file_name"] = name
        layer_attributes = {"name": name, "metadata": meta}

        # TODO:  dump metadata dictionary. json or pickle?

        # return [(data, layer_attributes, layer_type)]  # (data,meta) is fine since 'image' is default
        # JAH: refactor to make channel_index be a zslice

        print(f"loaded {name} size: {data.shape}")
        if self._channel_index < 0:
            return data
        else:
            return data[:, self._channel_index, :, :]

    def _format_output(self, image: np.ndarray):
        """
        Format segmented images to uint8 to save via AICSImage

        Params:
            image (np.ndarray): segmented image

        Returns
            np.ndarray: image converted to uint8 for saving if boolean... otherwise
        """
        if image.dtype == "bool" or image.dtype == np.uint8:
            image = image.astype(np.uint16)
            image[image > 0] = 1
            msg = f"converted boolean to {image.dtype}. "
            self._write_to_log_file(msg)
        else:
            image = image.astype(np.uint16)
            msg = f" enforced  {image.dtype}"
            print(msg)
            self._write_to_log_file(msg)

        return image

    def _write_to_log_file(self, text: str):
        with open(self._log_path, "a") as writer:
            writer.write(f"{text}\n")

    def _get_input_files(self, input_dir: Path, extensions: List[str]) -> List[Path]:
        input_files = list()
        for ext in extensions:
            input_files.extend(input_dir.glob(f"*.{ext}"))
        return input_files



from enum import Enum
from dataclasses import dataclass
from typing import Dict, List, Union


class WidgetType(Enum):
    SLIDER = "slider"
    DROPDOWN = "drop-down"

    @staticmethod
    def from_str(value: str):
        if value is not None:
            value = value.lower()
        if value == WidgetType.SLIDER.value:
            return WidgetType.SLIDER
        if value == WidgetType.DROPDOWN.value:
            return WidgetType.DROPDOWN
        raise NotImplementedError()


@dataclass
class FunctionParameter:
    """
    Represents an input parameter to a segmenter function
    """

    name: str
    widget_type: WidgetType
    data_type: str
    min_value: Union[int, float] = None
    max_value: Union[int, float] = None
    increment: Union[int, float] = None
    options: List[str] = None


@dataclass
class SegmenterFunction:
    """
    Represents an aicssegmentation function.
    Functions are the smallest executable entity in a workflow and directly map
    to a python callable
    """

    name: str
    display_name: str
    function: str
    module: str
    parameters: Dict[str, List[FunctionParameter]] = None



import importlib
import numpy as np

from dataclasses import dataclass
from enum import Enum
from typing import Dict, List, Any
from infer_subc.workflow.segmenter_function import SegmenterFunction


class WorkflowStepCategory(Enum):
    EXTRACTION = "extraction"
    PRE_PROCESSING = "preprocessing"
    CORE = "core"
    POST_PROCESSING = "postprocessing"
    POST_POST_PROCESSING = "postpostprocessing"
    EXPORT = "export"

    @staticmethod
    def from_str(value: str):
        if value is not None:
            value = value.lower()
        if value == WorkflowStepCategory.EXTRACTION.value:
            return WorkflowStepCategory.EXTRACTION
        if value == WorkflowStepCategory.PRE_PROCESSING.value:
            return WorkflowStepCategory.PRE_PROCESSING
        if value == WorkflowStepCategory.CORE.value:
            return WorkflowStepCategory.CORE
        if value == WorkflowStepCategory.POST_PROCESSING.value:
            return WorkflowStepCategory.POST_PROCESSING
        if value == WorkflowStepCategory.POST_POST_PROCESSING.value:
            return WorkflowStepCategory.POST_POST_PROCESSING
        if value == WorkflowStepCategory.EXPORT.value:
            return WorkflowStepCategory.EXPORT
        raise NotImplementedError()


@dataclass
class WorkflowStep:
    """
    Represents a single step in an aicssegmentation Workflow
    """

    category: WorkflowStepCategory
    function: SegmenterFunction
    step_number: int
    parent: List[int]
    parameter_values: Dict[str, List] = None

    @property
    def name(self):
        return self.function.display_name

    def execute(self, input_images: List[np.ndarray], parameters: Dict[str, Any] = None) -> np.ndarray:
        """
        Execute this workflow step on the given input image and return the result.

        Params:
            input_images (List[np.ndarray]): List of image inputs to perform this
                    workflow step on, generally parent image
            parameters (Dict): Dictionary of parameters to pass to the
                                underlying function

        Returns
            self.result (np.ndarray): Result of performing workflow step
                                        on the given image.
        """
        if not isinstance(input_images, list):
            raise ValueError("input_images must be a list")

        if parameters is not None and not self._check_parameters(parameters):
            raise ValueError(
                "Provided parameters are invalid. All keys in the parameters dictionary"
                "must correspond to existing parameter names defined for the underlying workflow function."
                "Note: parameter names are case sensitive"
            )

        py_module = importlib.import_module(self.function.module)
        py_function = getattr(py_module, self.function.function)

        try:
            # Most functions require unpacking the images
            if parameters is not None:
                return py_function(*input_images, **parameters)

            return py_function(*input_images)
        except TypeError:
            # Some functions want it as a list
            if parameters is not None:
                return py_function(input_images, **parameters)
            return py_function(input_images)

    def _check_parameters(self, parameters: Dict[str, Any]) -> bool:
        for key in parameters.keys():
            if key not in self.function.parameters.keys():
                return False

        return True



from functools import partial
from logging import getLogger
from pathlib import Path
from typing import Any, Dict, List, Optional, Union, Sequence, Tuple, Callable
import numpy as np
import xarray as xr

from aicsimageio import AICSImage, exceptions
from aicsimageio.dimensions import DimensionNames

# # TODO: need to depricate these
# from qtpy.QtWidgets import (
#     QCheckBox,
#     QGroupBox,
#     QListWidget,
#     QListWidgetItem,
#     QVBoxLayout,
# )
from aicsimageio.writers import OmeTiffWriter

FullLayerData = Tuple[Any, Dict, str]
LayerData = Union[Tuple[Any], Tuple[Any, Dict], FullLayerData]

PathLike = Union[str, Path]
# PathOrPaths = Union[str, Sequence[str]]
PathOrPaths = Union[PathLike, Sequence[PathLike]]
# ReaderFunction = Callable[[PathOrPaths], List[LayerData]]
# from napari.types import LayerData, PathLike, ReaderFunction

logger = getLogger(__name__)

###############################################################################
AICSIMAGEIO_CHOICES = "AICSImageIO Scene Management"
CLEAR_LAYERS_ON_SELECT = "Clear All Layers on New Scene Selection"
UNPACK_CHANNELS_TO_LAYERS = "Unpack Channels as Layers"
SCENE_LABEL_DELIMITER = " :: "
# Threshold above which to use out-of-memory loading
IN_MEM_THRESHOLD_PERCENT = 0.3
IN_MEM_THRESHOLD_SIZE_BYTES = 4e9  # 4GB
###############################################################################


# from aicsimageio.writers import OmeTiffWriter
# from napari_aicsimageio.core import reader_function


########## .   napari-aicsimageio functions..
def _get_full_image_data(
    img: AICSImage,
    in_memory: bool,
) -> xr.DataArray:
    if DimensionNames.MosaicTile in img.reader.dims.order:
        try:
            if in_memory:
                return img.reader.mosaic_xarray_data.squeeze()

            return img.reader.mosaic_xarray_dask_data.squeeze()

        # Catch reader does not support tile stitching
        except NotImplementedError:
            logger.warning("AICSImageIO: Mosaic tile stitching " "not yet supported for this file format reader.")

    if in_memory:
        return img.reader.xarray_data.squeeze()

    return img.reader.xarray_dask_data.squeeze()


# Function to get Metadata to provide with data
def _get_meta(path: "PathLike", data: xr.DataArray, img: AICSImage) -> Dict[str, Any]:
    meta: Dict[str, Any] = {}
    if DimensionNames.Channel in data.dims:
        # Construct basic metadata
        # Use filename if single scene and no scene name is available
        if len(img.scenes) == 1 and img.current_scene == "Image:0":
            channels_with_scene_index = [
                f"{Path(path).stem}{SCENE_LABEL_DELIMITER}{channel_name}"
                for channel_name in data.coords[DimensionNames.Channel].data.tolist()
            ]
        else:
            channels_with_scene_index = [
                f"{img.current_scene_index}{SCENE_LABEL_DELIMITER}"
                f"{img.current_scene}{SCENE_LABEL_DELIMITER}{channel_name}"
                for channel_name in data.coords[DimensionNames.Channel].data.tolist()
            ]
        meta["name"] = channels_with_scene_index
        meta["channel_axis"] = data.dims.index(DimensionNames.Channel)

    # Not multi-channel, use current scene as image name
    else:
        # use filename if single scene and no scene name is available
        if len(img.scenes) == 1 and img.current_scene == "Image:0":
            meta["name"] = Path(path).stem
        else:
            meta["name"] = img.reader.current_scene

    # Handle samples / RGB
    if DimensionNames.Samples in img.reader.dims.order:
        meta["rgb"] = True

    # Handle scales
    scale: List[float] = []
    # check the dims of the squeezed array for scale values
    for dim in data.dims:
        if dim in [
            DimensionNames.SpatialX,
            DimensionNames.SpatialY,
            DimensionNames.SpatialZ,
        ]:
            scale_val = getattr(img.physical_pixel_sizes, dim)
            if scale_val is not None:
                scale.append(scale_val)

    # Apply scales
    if len(scale) > 0:
        meta["scale"] = tuple(scale)

    # Apply all other metadata
    img_meta = {"aicsimage": img, "raw_image_metadata": img.metadata}
    try:
        img_meta["ome_types"] = img.ome_metadata
    except Exception:
        pass

    meta["metadata"] = img_meta
    return meta


# DEPRICATED:  not using in napari context
# def _widget_is_checked(widget_name: str) -> bool:
#     import napari

#     # Get napari viewer from current process
#     viewer = napari.current_viewer()

#     # Get scene management widget
#     scene_manager_choices_widget = viewer.window._dock_widgets[AICSIMAGEIO_CHOICES]
#     for child in scene_manager_choices_widget.widget().children():
#         if isinstance(child, QCheckBox):
#             if child.text() == widget_name:
#                 return child.isChecked()

#     return False

# DEPRICATED:  not using in napari context
# # Function to handle multi-scene files.
# def _get_scenes(path: "PathLike", img: AICSImage, in_memory: bool) -> None:
#     import napari

#     # Get napari viewer from current process
#     viewer = napari.current_viewer()

#     # Add a checkbox widget if not present
#     if AICSIMAGEIO_CHOICES not in viewer.window._dock_widgets:
#         # Create a checkbox widget to set "Clear On Scene Select" or not
#         scene_clear_checkbox = QCheckBox(CLEAR_LAYERS_ON_SELECT)
#         scene_clear_checkbox.setChecked(False)

#         # Create a checkbox widget to set "Unpack Channels" or not
#         channel_unpack_checkbox = QCheckBox(UNPACK_CHANNELS_TO_LAYERS)
#         channel_unpack_checkbox.setChecked(False)

#         # Add all scene management state to a single box
#         scene_manager_group = QGroupBox()
#         scene_manager_group_layout = QVBoxLayout()
#         scene_manager_group_layout.addWidget(scene_clear_checkbox)
#         scene_manager_group_layout.addWidget(channel_unpack_checkbox)
#         scene_manager_group.setLayout(scene_manager_group_layout)
#         scene_manager_group.setFixedHeight(100)

#         viewer.window.add_dock_widget(
#             scene_manager_group,
#             area="right",
#             name=AICSIMAGEIO_CHOICES,
#         )

#     # Create the list widget and populate with the ids & scenes in the file
#     list_widget = QListWidget()
#     for i, scene in enumerate(img.scenes):
#         list_widget.addItem(f"{i}{SCENE_LABEL_DELIMITER}{scene}")

#     # Add this files scenes widget to viewer
#     viewer.window.add_dock_widget(
#         list_widget,
#         area="right",
#         name=f"{Path(path).name}{SCENE_LABEL_DELIMITER}Scenes",
#     )

#     # Function to create image layer from a scene selected in the list widget
#     def open_scene(item: QListWidgetItem) -> None:
#         scene_text = item.text()

#         # Use scene indexes to cover for duplicate names
#         scene_index = int(scene_text.split(SCENE_LABEL_DELIMITER)[0])

#         # Update scene on image and get data
#         img.set_scene(scene_index)
#         data = _get_full_image_data(img=img, in_memory=in_memory)

#         # Get metadata and add to image
#         meta = _get_meta("", data, img)

#         # Optionally clear layers
#         if _widget_is_checked(CLEAR_LAYERS_ON_SELECT):
#             viewer.layers.clear()

#         # Optionally remove channel axis
#         if not _widget_is_checked(UNPACK_CHANNELS_TO_LAYERS):
#             meta["name"] = scene_text
#             meta.pop("channel_axis", None)

#         viewer.add_image(data, **meta)

#     list_widget.currentItemChanged.connect(open_scene)  # type: ignore


def reader_function(path: "PathLike", in_memory: Optional[bool] = None) -> Optional[List["LayerData"]]:
    """
    Given a single path return a list of LayerData tuples.
    """
    # Only support single path
    if isinstance(path, list):
        logger.info("AICSImageIO: Multi-file reading not yet supported.")
        return None

    # TODO:  remove these imports?  is it slowing things down?
    if in_memory is None:
        from aicsimageio.utils.io_utils import pathlike_to_fs
        from psutil import virtual_memory

        fs, path = pathlike_to_fs(path)
        imsize = fs.size(path)
        available_mem = virtual_memory().available
        _in_memory = imsize <= IN_MEM_THRESHOLD_SIZE_BYTES and imsize / available_mem <= IN_MEM_THRESHOLD_PERCENT
    else:
        _in_memory = in_memory

    # Alert console of how we are loading the image
    logger.info(f"AICSImageIO: Reader will load image in-memory: {_in_memory}")

    # Open file and get data
    img = AICSImage(path)

    # Check for multiple scenes
    if len(img.scenes) > 1:
        logger.info(
            f"AICSImageIO: Image contains {len(img.scenes)} scenes. "
            f"Supporting more than the first scene is experimental. "
            f"Select a scene from the list widget. There may be dragons!"
        )
        # # Launch the list widget
        # # DEPRICATED:  no napari context... scenes unsupported
        # _get_scenes(path=path, img=img, in_memory=_in_memory)

        # Return an empty LayerData list; ImgLayers will be handled via the widget.
        # HT Jonas Windhager
        return [(None,)]
    else:
        data = _get_full_image_data(img, in_memory=_in_memory)
        meta = _get_meta(path, data, img)
        return [(data.data, meta, "image")]


# def get_reader(
#     path: "PathLike", in_memory: Optional[bool] = None
# ) -> Optional["ReaderFunction"]:
#     """
#     Given a single path or list of paths, return the appropriate aicsimageio reader.
#     """
#     # Only support single path
#     if isinstance(path, list):
#         logger.info("AICSImageIO: Multi-file reading not yet supported.")
#         return None

#     # See if there is a supported reader for the file(s) provided
#     try:
#         # There is an assumption that the images are stackable and
#         # I think it is also safe to assume that if stackable, they are of the same type
#         # So only determine reader for the first one
#         AICSImage.determine_reader(path)

#         # The above line didn't error so we know we have a supported reader
#         # Return a partial function with in_memory determined
#         return partial(reader_function, in_memory=in_memory)

#     # No supported reader, return None
#     except exceptions.UnsupportedFileFormatError:
#         logger.warning("AICSImageIO: Unsupported file format.")
#         return None

#     except Exception as e:
#         logger.warning("AICSImageIO: exception occurred during reading...")
#         logger.warning(e)
#         logger.warning(
#             "If this issue looks like a problem with AICSImageIO, "
#             "please file a bug report: "
#             "https://github.com/AllenCellModeling/napari-aicsimageio"
#         )
#         return None


def export_ome_tiff(data_in, meta_in, img_name, out_path, channel_names) -> str:
    """
    wrapper for exporting ome tiff

    #  data_in: types.ArrayLike,
    #  meta_in: dict,
    # img_name: types.PathLike,
    # out_path: types.PathLike,
    # curr_chan: int
    # assumes a single image
    """

    # TODO:  figure out if the meta_in["metadata"]['aicsimage] is passed on?  .ome.tiffs are x2 size
    out_name = out_path + img_name + ".ome.tiff"

    image_names = [img_name]
    # chan_names = meta_in['metadata']['aicsimage'].channel_names

    physical_pixel_sizes = [meta_in["metadata"]["aicsimage"].physical_pixel_sizes]

    dimension_order = ["CZYX"]
    if channel_names is None:
        channel_names = [meta_in["metadata"]["aicsimage"].channel_names]
    else:
        channel_names = [channel_names]

    if len(data_in.shape) == 3:  # single channel zstack
        data_in = data_in[np.newaxis, :, :, :]
    elif len(data_in.shape) == 2:  # single channel , 1Z
        data_in = data_in[np.newaxis, np.newaxis, :, :]

    if data_in.dtype == "bool":
        data_in = data_in.astype(np.uint16)
        data_in[data_in > 0] = 1

    out_ome = OmeTiffWriter.build_ome(
        [data_in.shape],
        [data_in.dtype],
        channel_names=channel_names,  # type: ignore
        image_name=image_names,
        physical_pixel_sizes=physical_pixel_sizes,
        dimension_order=dimension_order,
    )

    OmeTiffWriter.save(
        data_in,
        out_name,
        dim_order=dimension_order,
        channel_names=channel_names,
        image_names=image_names,
        physical_pixel_sizes=physical_pixel_sizes,
        ome_xml=out_ome,
    )
    return out_name



# from aicssegmentation
from pathlib import Path
from typing import Union


class FileSystemUtilities:
    @staticmethod
    def create_directory(path: Union[Path, str]):
        """
        Create directory for the given path. This will create all parent directories as needed.
        """
        Path(path).mkdir(exist_ok=True, parents=True)



import numpy as np
from typing import Any, List, Union
from pathlib import Path

from infer_subc.core.img import apply_mask

import pandas as pd

from infer_subc.utils.stats import ( get_aXb_stats_3D, 
                    get_summary_stats_3D, 
                    get_simple_stats_3D, 
                    get_radial_stats, 
                    get_depth_stats, 
                    _assert_uint16_labels )

from infer_subc.utils.batch import list_image_files, find_segmentation_tiff_files
from infer_subc.core.file_io import read_czi_image, read_tiff_image

from infer_subc.constants import organelle_to_colname, NUC_CH, GOLGI_CH, PEROX_CH

def make_organelle_stat_tables(
    organelle_names: List[str],
    organelles: List[np.ndarray],
    intensities: List[np.ndarray],
    nuclei_obj:np.ndarray, 
    cellmask_obj:np.ndarray,
    organelle_mask: np.ndarray, 
    out_data_path: Path, 
    source_file: str,
    n_rad_bins: Union[int,None] = None,
    n_zernike: Union[int,None] = None,
) -> int:
    """
    get summary and all cross stats between organelles `a` and `b`
    calls `get_summary_stats_3D`
    """
    count = 0
    org_stats_tabs = []
    for j, target in enumerate(organelle_names):
        org_img = intensities[j]        
        org_obj = _assert_uint16_labels(organelles[j])

        # A_stats_tab, rp = get_simple_stats_3D(A,mask)
        a_stats_tab, rp = get_summary_stats_3D(org_obj, org_img, organelle_mask)
        a_stats_tab.insert(loc=0,column='organelle',value=target )
        a_stats_tab.insert(loc=0,column='ID',value=source_file.stem )

        # add the touches for all other organelles
        # loop over Bs
        merged_tabs = []
        for i, nmi in enumerate(organelle_names):
            if i != j:
                # get overall stats of intersection
                # print(f"  b = {nmi}")
                count += 1
                # add the list of touches
                b = _assert_uint16_labels(organelles[i])

                ov = []
                b_labs = []
                labs = []
                for idx, lab in enumerate(a_stats_tab["label"]):  # loop over A_objects
                    xyz = tuple(rp[idx].coords.T)
                    cmp_org = b[xyz]
                    
                    # total number of overlapping pixels
                    overlap = sum(cmp_org > 0)
                    # overlap?
                    labs_b = cmp_org[cmp_org > 0]
                    b_js = np.unique(labs_b).tolist()

                    # if overlap > 0:
                    labs.append(lab)
                    ov.append(overlap)
                    b_labs.append(b_js)

                cname = organelle_to_colname[nmi]
                # add organelle B columns to A_stats_tab
                a_stats_tab[f"{cname}_overlap"] = ov
                a_stats_tab[f"{cname}_labels"] = b_labs  # might want to make this easier for parsing later

                #####  2  ###########
                # get cross_stats

                cross_tab = get_aXb_stats_3D(org_obj, b, organelle_mask) 
                shell_cross_tab = get_aXb_stats_3D(org_obj, b, organelle_mask, use_shell_a=True)
                            
                # cross_tab["organelle_b"]=nmi
                # shell_cross_tab["organelle_b"]=nmi
                #  Merge cross_tabs and shell_cross_tabs 
                # merged_tab = pd.merge(cross_tab,shell_cross_tab, on="label_")
                merged_tab = pd.concat([cross_tab,shell_cross_tab])
                merged_tab.insert(loc=0,column='organelle_b',value=nmi )

                merged_tabs.append( merged_tab )


        #  Now append the 
        # csv_path = out_data_path / f"{source_file.stem}-{target}_shellX{nmi}-stats.csv"
        # e_stats_tab.to_csv(csv_path)
        # stack these tables for each organelle
        crossed_tab = pd.concat(merged_tabs)
        # csv_path = out_data_path / f"{source_file.stem}-{target}X{nmi}-stats.csv"
        # stats_tab.to_csv(csv_path)
        crossed_tab.insert(loc=0,column='organelle',value=target )
        crossed_tab.insert(loc=0,column='ID',value=source_file.stem )

        # now get radial stats
        rad_stats,z_stats, _ = get_radial_stats(        
                cellmask_obj,
                organelle_mask,
                org_obj,
                org_img,
                target,
                nuclei_obj,
                n_rad_bins,
                n_zernike
                )

        d_stats = get_depth_stats(        
                cellmask_obj,
                organelle_mask,
                org_obj,
                org_img,
                target,
                nuclei_obj
                )
      
        proj_stats = pd.merge(rad_stats, z_stats,on=["organelle","mask"])
        proj_stats = pd.merge(proj_stats, d_stats,on=["organelle","mask"])
        proj_stats.insert(loc=0,column='ID',value=source_file.stem )


        # write out files... 
        # org_stats_tabs.append(A_stats_tab)
        csv_path = out_data_path / f"{source_file.stem}-{target}-stats.csv"
        a_stats_tab.to_csv(csv_path)

        csv_path = out_data_path / f"{source_file.stem}-{target}-cross-stats.csv"
        crossed_tab.to_csv(csv_path)

        csv_path = out_data_path / f"{source_file.stem}-{target}-proj-stats.csv"
        proj_stats.to_csv(csv_path)

        count += 1

    print(f"dumped {count}x3 organelle stats ({organelle_names}) csvs")
    return count


def dump_all_stats_tables(int_path: Union[Path,str], 
                   out_path: Union[Path, str], 
                   raw_path: Union[Path,str], 
                   organelle_names: List[str]= ["nuclei","golgi","peroxi"], 
                   organelle_chs: List[int]= [NUC_CH,GOLGI_CH, PEROX_CH], 
                    ) -> int :
    """  
    TODO: add loggin instead of printing
        append tiffcomments with provenance
    """

    
    if isinstance(raw_path, str): raw_path = Path(raw_path)
    if isinstance(int_path, str): int_path = Path(int_path)
    if isinstance(out_path, str): out_path = Path(out_path)
    
    img_file_list = list_image_files(raw_path,".czi")

    if not Path.exists(out_path):
        Path.mkdir(out_path)
        print(f"making {out_path}")
        
    for img_f in img_file_list:
        filez = find_segmentation_tiff_files(img_f, organelle_names, int_path)
        img_data,meta_dict = read_czi_image(filez["raw"])

        # load organelles and masks
        cyto_mask = read_tiff_image(filez["cyto"])
        cellmask_obj = read_tiff_image(filez["cell"])



        # create intensities from raw as list
        intensities = [img_data[ch] for ch in organelle_chs]

        # load organelles as list
        organelles = [read_tiff_image(filez[org]) for org in organelle_names]
        
        #get mask (cyto_mask)
        nuclei_obj = organelles[ organelle_names.index("nuc") ]

        n_files = make_organelle_stat_tables(organelle_names, 
                                      organelles,
                                      intensities, 
                                      nuclei_obj,
                                      cellmask_obj,
                                      cyto_mask, 
                                      out_path, 
                                      img_f,
                                      n_rad_bins=5,
                                      n_zernike=9)

    return n_files



# These are probably not to be used
def dump_shell_cross_stats(
    organelle_names: List[str], organelles: List[np.ndarray], mask: np.ndarray, out_data_path: Path, source_file: str
) -> int:
    """
    get all cross stats between organelles `a` and `b`, and "shell of `a`" and `b`.   "shell" is the boundary of `a`
    calls `get_aXb_stats_3D`
    """
    count = 0
    for j, target in enumerate(organelle_names):
        # print(f"getting stats for org_obj = {target}")
        org_obj = organelles[j]
        # loop over Bs
        for i, nmi in enumerate(organelle_names):
            if i != j:
                # get overall stats of intersection
                # print(f"  X {nmi}")
                b = organelles[i]
                stats_tab = get_aXb_stats_3D(org_obj, b, mask)
                csv_path = out_data_path / f"{source_file.stem}-{target}X{nmi}-stats.csv"
                stats_tab.to_csv(csv_path)

                e_stats_tab = get_aXb_stats_3D(org_obj, b, mask, use_shell_a=True)
                csv_path = out_data_path / f"{source_file.stem}-{target}_shellX{nmi}-stats.csv"
                e_stats_tab.to_csv(csv_path)

                count += 1
    print(f"dumped {count} x2 organelle cross-stats csvs")
    return count





def load_summary_stats_csv(in_path: Path) -> pd.DataFrame:
    """ helper to load the summary stats csv: summary-stats.csv
    returns pandas DataFrame """
    csv_path = in_path / f"summary-stats.csv"
    summary_df = pd.read_csv(csv_path, index_col=0)
    # need to convert columns *_labels
    list_cols = [col for col in summary_df.columns if "labels" in col] #if col.contains("label")
    summary_df = fix_int_list_cols(summary_df,list_cols)
    return summary_df


def load_summary_proj_stats_csv(in_path: Path) -> pd.DataFrame:
    """ helper to load summary projection stats csv: summary-proj-stats.csv
    returns pandas DataFrame """
    obj_cols =  ['ID', 'organelle','mask','radial_n_bins','n_z']  # leave alone
    str_cols = [ 'radial_bins']
    int_cols = ['radial_cm_vox_cnt', 'radial_org_vox_cnt', 'radial_org_intensity', 'radial_n_pix','zernike_n', 'zernike_m', 'z','z_cm_vox_cnt','z_org_vox_cnt', 'z_org_intensity', 'z_nuc_vox_cnt']
    float_cols = ['radial_cm_cv', 'radial_org_cv', 'radial_img_cv','zernike_cm_mag', 'zernike_cm_phs','zernike_obj_mag', 'zernike_obj_phs', 'zernike_nuc_mag','zernike_nuc_phs', 'zernike_img_mag']

    csv_path = in_path / f"summary-proj-stats.csv"
    proj = pd.read_csv(csv_path, index_col=0)
    proj = fix_str_list_cols(proj, str_cols)
    proj = fix_int_list_cols(proj, int_cols)
    proj = fix_float_list_cols(proj, float_cols)
    return proj
        

def load_summary_cross_stats_csv(in_path: Path) -> pd.DataFrame:
    """ helper to load summary cross- stats csv: summary-cross-stats.csv
    returns pandas DataFrame """

    csv_path = in_path / f"summary-cross-stats.csv"
    summary_df = pd.read_csv(csv_path, index_col=0)

    list_cols = [col for col in summary_df.columns if "label" in col] #if col.contains("label")
    str_list_cols = [col for col in list_cols if "__" in col]
    int_list_cols = [col for col in list_cols if "__" not in col]

    summary_df = fix_str_list_cols(summary_df,str_list_cols)
    summary_df = fix_int_list_cols(summary_df,int_list_cols)

    return summary_df
    


# for a list of "prefixes"  collect stats + cross stats masked by cytosol (including nuclei masked by cellmask)

def summarize_by_id(stats_in:pd.DataFrame,agg_fn: List) -> pd.DataFrame:
    """ 
    """
    summary = stats_in.groupby(['ID']).agg(agg_fn)
    summary.columns = ["_".join(col_name).rstrip('_') for col_name in summary.columns.to_flat_index()]
    return summary



def create_stats_summary(summary_df:pd.DataFrame) -> pd.DataFrame:
    """
    """
    column_names = summary_df.columns

    def frac(x):
        return (x>0).sum()/x.count() 

    math_cols = ['ID', 'mean_intensity',
        'standard_deviation_intensity',
        'min_intensity','max_intensity', 'equivalent_diameter',
        'euler_number', 'extent']
    vol_cols = ['ID','volume']
    overlap_cols = ['ID'] + [col for col in column_names if col.endswith('_overlap')]
    labels_cols = ['ID'] + [col for col in column_names if col.endswith('_labels')]
   
    agg_func_math = ['sum', 'mean', 'median', 'min', 'max', 'std','count']
    agg_func_overlap = ['sum', 'mean', 'median','count',frac]
    agg_func_labels = ['sum']
    agg_func_vol = ['sum', 'mean', 'median', 'min', 'max', 'std', 'var']

    math_summary = summarize_by_id( summary_df[math_cols] , agg_func_math)
    
    # label_stats = fix_list_col(summary_df[labels_cols])
    label_summary = summarize_by_id( summary_df[labels_cols] , agg_func_labels)
    overlap_summary = summarize_by_id( summary_df[overlap_cols] ,agg_func_overlap)
    vol_summary = summarize_by_id( summary_df[vol_cols] , agg_func_vol)
    result = pd.concat([math_summary, vol_summary, overlap_summary, label_summary], axis=1)

    result.insert(loc=0,column="ID",value=result.index)

    return result


def summarize_by_group(stats_in:pd.DataFrame, grp_col:list, agg_fn:list) -> pd.DataFrame:
    """ 
    """
    summary = stats_in.reset_index(drop=True).groupby(grp_col).agg(agg_fn)
    summary.columns = ["_".join(col_name).rstrip('_') for col_name in summary.columns.to_flat_index()]
    return summary


def create_cross_stats_summary(summary_df:pd.DataFrame) -> pd.DataFrame:
    """
    """
    # dropped_cols = ['centroid-0', 'centroid-1', 'centroid-2', 'bbox-0', 'bbox-1', 'bbox-2', 'bbox-3', 'bbox-4', 'bbox-5']
    # cross_cols = ['ID', 'organelle', 'organelle_b', 'shell', 'label_', 'label', 'volume',
    #    'equivalent_diameter','surface_area', 'label_a', 'label_b']

    group_cols = ['ID','organelle_b', 'shell']
    id_cols = ['label_','label_a', 'label_b'] 
    math_cols = ['volume','equivalent_diameter','surface_area']

    def lst(x):
        return x.to_list()
       
    agg_func_math = ['sum', 'mean', 'median', 'min', 'max', 'std','count']
    agg_func_id = [lst]

    math_summary = summarize_by_group( summary_df[group_cols + math_cols],group_cols, agg_func_math)

    id_summary = summarize_by_group( summary_df[group_cols + id_cols],group_cols, agg_func_id)

    result = pd.concat([math_summary, id_summary], axis=1)
    return result

    # now 

def summarize_cross_stats(summary_df:pd.DataFrame) -> pd.DataFrame:
    """
    """
    # get shell
    shell_summary_i = create_cross_stats_summary(summary_df.loc[summary_df["shell"] == True]).reset_index().drop("shell", axis = 1).add_prefix("shell_")
    # rename shell_ID to ID
    shell_summary_i = shell_summary_i.rename(columns={"shell_ID":"ID","shell_organelle_b":"organelle_b"})

    # get non-shell
    summary_i = create_cross_stats_summary(summary_df.loc[summary_df["shell"] == False]).reset_index().drop("shell", axis = 1)

    summary_i = summary_i.merge(shell_summary_i,on=["ID","organelle_b"])

    return summary_i


def pivot_cross_stats(summary_df:pd.DataFrame) -> pd.DataFrame:
    """
    """
    xstat_df = pd.DataFrame()
    org_bs = summary_df["organelle_b"].unique()
    for i,org_b in enumerate(org_bs):
        org_i = summary_df.loc[summary_df["organelle_b"] == org_b]

        # get shell
        shell_summary_i = create_cross_stats_summary(org_i.loc[org_i["shell"] == False]).reset_index().drop("shell", axis = 1).add_prefix("shell_")
        # rename shell_ID to ID
        shell_summary_i = shell_summary_i.rename(columns={"shell_ID":"ID","shell_organelle_b":"organelle_b"})
        # get non-shell
        summary_i = create_cross_stats_summary(org_i.loc[org_i["shell"] == False]).reset_index().drop("shell", axis = 1)
        col_orgb = organelle_to_colname[org_b]

        summary_i = summary_i.merge(shell_summary_i,on=["ID","organelle_b"]).drop("organelle_b", axis=1).add_suffix(f"_{col_orgb}")
        if i>0:
            xstat_df = pd.concat([xstat_df,summary_i], axis=1)
        else:
            xstat_df = summary_i
            
    id_cols = [col for col in xstat_df.columns if "ID" in col]
    IDcol = xstat_df[id_cols[0]]
    xstat_df = xstat_df.drop(id_cols, axis=1)
    xstat_df.insert(loc=0,column="ID",value=IDcol)

    return xstat_df


def fix_int_list_cols(in_df:pd.DataFrame, list_cols) -> pd.DataFrame:
    """ 
    """

    def _str_to_list(x):
        if x == '[]':
            return list()
        elif isinstance(x,float): # catch nans
            return x
        else:
            xstr = x.strip("[]").replace("'", "").split(", ")
        return [int(float(x)) for x in xstr]
        
    out_df = pd.DataFrame() 
    for col in in_df.columns:    
        out_df[col] = in_df[col].apply(_str_to_list) if col in list_cols else in_df[col]
    return out_df

def fix_float_list_cols(in_df:pd.DataFrame, list_cols) -> pd.DataFrame:
    """ 
    """
    def _str_to_list(x):
        if x == '[]':
            return list()
        elif isinstance(x,float): # catch nans
            return x
        else:
            xstr = x.strip("[]").replace("'", "").split(", ")
        return [float(x) for x in xstr]
        
    out_df = pd.DataFrame() 
    for col in in_df.columns:    
        out_df[col] = in_df[col].apply(_str_to_list) if col in list_cols else in_df[col]
    return out_df

def fix_str_list_cols(in_df:pd.DataFrame, list_cols) -> pd.DataFrame:
    """ 
    """
    def _str_to_list(x):
        if x == '[]':
            return list()
        elif isinstance(x,float): # catch nans
            return x
        else:
            xstr = x.strip("[]").replace("'", "").split(", ")
        return [x for x in xstr]
        
    out_df = pd.DataFrame() 
    for col in in_df.columns:    
        out_df[col] = in_df[col].apply(_str_to_list) if col in list_cols else in_df[col]
    return out_df


def load_stats_csv(in_path: Path, img_id: str, target_org: str) -> pd.DataFrame:
    """ helper to load the basic stats csv: `img_id`-`target_organelle` -stats.csv
    returns pandas DataFrame """
    csv_path = in_path / f"{img_id}-{target_org}-stats.csv"
    stats = pd.read_csv(csv_path, index_col=0,dtype={"ID":str,"organelle":str})
    # need to convert columns *_labels
    list_cols = [col for col in stats.columns if col.endswith('_labels')]
    stats = fix_int_list_cols(stats,list_cols)
    return stats
        

def load_proj_stats_csv(in_path: Path, img_id: str, target_org: str) -> pd.DataFrame:
    """ helper to load  the projection stats csv: `img_id`-`target_organelle` -proj-stats.csv
    returns pandas DataFrame """
    # obj_cols =  ['ID', 'organelle','radial_n_bins','n_z']  # leave alone
    # str_cols = [ 'radial_bins']
    int_cols = ['radial_cm_vox_cnt', 'radial_org_vox_cnt', 'radial_org_intensity', 'radial_n_pix','zernike_n', 'zernike_m', 'z','z_cm_vox_cnt','z_org_vox_cnt', 'z_org_intensity', 'z_nuc_vox_cnt']
    float_cols = ['radial_cm_cv', 'radial_org_cv', 'radial_img_cv','zernike_cm_mag', 'zernike_cm_phs','zernike_obj_mag', 'zernike_obj_phs', 'zernike_nuc_mag','zernike_nuc_phs', 'zernike_img_mag']

    csv_path = in_path / f"{img_id}-{target_org}-proj-stats.csv"
    proj = pd.read_csv(csv_path, index_col=0)
    proj['radial_bins'] = proj['radial_bins'].values.squeeze().tolist()
    # proj = fix_str_list_cols(proj, str_cols)
    proj = fix_int_list_cols(proj, int_cols)
    proj = fix_float_list_cols(proj, float_cols)
    return proj
        

def load_cross_stats_csv(in_path: Path, img_id: str, target_org: str) -> pd.DataFrame:
    """ helper to load  the cross- stats csv: `img_id`-`target_organelle` -cross-stats.csv
    returns pandas DataFrame """
    csv_path = in_path / f"{img_id}-{target_org}-cross-stats.csv"
    cross = pd.read_csv(csv_path, index_col=0)
    return cross


def summarize_organelle_stats(int_path: Union[Path,str], 
                              organelle_names: List[str]= ["nuclei","golgi","peroxi"]):
    """  
    """
    # write out files... 

    if isinstance(int_path, str): int_path = Path(int_path)


    all_stats_df = pd.DataFrame()
    all_cross_stats_df = pd.DataFrame()
    all_proj_stats_df = pd.DataFrame()
    
    for target in organelle_names:
        stat_file_list = sorted( int_path.glob(f"*{target}-stats.csv") )

        stats_df = pd.DataFrame()
        cross_stats_df = pd.DataFrame()
        proj_stats_df = pd.DataFrame()

        for stats_f in stat_file_list:
            stem = stats_f.stem.split("-")[0]
            # stats load the csv
            stats = load_stats_csv(int_path,stem, target)
            # projection stats
            proj = load_proj_stats_csv(int_path,stem, target)
            # cross stats
            cross = load_cross_stats_csv(int_path,stem, target)

            stats_df = pd.concat([stats_df,stats],axis=0, join='outer')
            proj_stats_df = pd.concat([proj_stats_df,proj],axis=0, join='outer')
            cross_stats_df = pd.concat([cross_stats_df,cross],axis=0, join='outer')
        

        ## maybe merge into all the possible files?
        # summary_df = pd.DataFrame(index=[f.stem.split("-")[0] for f in stat_file_list])
        # cross_stats_df = pd.DataFrame(index=[f.stem.split("-")[0] for f in stat_file_list])
        # proj_stats_df = pd.DataFrame(index=[f.stem.split("-")[0] for f in stat_file_list])

        summary_df = create_stats_summary(stats_df)
        summary_df.insert(loc=1,column="organelle",value=target)
        cross_summary_df = summarize_cross_stats(cross_stats_df)
        ## cross_summary_df = pivot_cross_stats(cross_stats_df)  #makes a wide version... but has a bug
        cross_summary_df.insert(loc=1,column="organelle",value=target)

        all_stats_df = pd.concat([all_stats_df,summary_df],axis=0)
        all_proj_stats_df = pd.concat([all_proj_stats_df,proj_stats_df],axis=0)
        all_cross_stats_df = pd.concat([all_cross_stats_df,cross_summary_df],axis=0)
    

    return all_stats_df, all_proj_stats_df, all_cross_stats_df
        



def dump_organelle_summary_tables(
                    int_path: Union[Path,str], 
                    out_path: Union[Path, str], 
                    organelle_names: List[str]= ["nuclei","golgi","peroxi"] ) -> int:
    """
    get summary and all cross stats between organelles `a` and `b`
    calls `get_summary_stats_3D`
    """

    if not Path.exists(out_path):
        Path.mkdir(out_path)
        print(f"making {out_path}")


    all_stats_df, all_proj_stats_df, all_cross_stats_df = summarize_organelle_stats( int_path, organelle_names )

    csv_path = out_path / f"summary-stats.csv"
    all_stats_df.to_csv(csv_path)

    csv_path = out_path / f"summary-proj-stats.csv"
    all_proj_stats_df.to_csv(csv_path)

    csv_path = out_path / f"summary-cross-stats.csv"
    all_cross_stats_df.to_csv(csv_path)

    return 1


def dump_stats(
    name: str,
    segmentation: np.ndarray,
    intensity_img: np.ndarray,
    mask: np.ndarray,
    out_data_path: Path,
    source_file: str,
) -> pd.DataFrame:
    """
    get summary stats of organelle only
    calls `get_summary_stats_3D`
    """
    stats_table, _ = get_summary_stats_3D(segmentation, intensity_img, mask)
    csv_path = out_data_path / f"{source_file.stem}-{name}-basic-stats.csv"
    stats_table.to_csv(csv_path)
    print(f"dumped {name} table to {csv_path}")

    return stats_table


# refactor to just to a target vs. list of probes
# for nuclei mask == cellmask
# for all oother mask == cytoplasm


def dump_projection_stats(
    organelle_names: List[str], 
    organelles: List[np.ndarray], 
    intensities: List[np.ndarray], 
    cellmask_obj:np.ndarray, 
    nuclei_obj:np.ndarray, 
    organelle_mask: np.ndarray, 
    out_data_path: Path, 
    source_file: str,
    n_rad_bins: Union[int,None] = None,
    n_zernike: Union[int,None] = None,
) -> int:
    """
    get all cross stats between organelles `a` and `b`, and "shell of `a`" and `b`.   "shell" is the boundary of `a`
    calls `get_proj_XYstats`  `get_proj_Zstats`
    """
    count = 0
    for j, organelle_name in enumerate(organelle_names):
       
        organelle_obj = organelles[j]
        organelle_img = intensities[j]

        rad_stats,z_stats, _ = get_radial_stats(        
                cellmask_obj,
                organelle_mask,
                organelle_obj,
                organelle_img,
                organelle_name,
                nuclei_obj,
                n_rad_bins,
                n_zernike
                )

        csv_path = out_data_path / f"{source_file.stem}-{organelle_name}-radial-stats.csv"
        rad_stats.to_csv(csv_path)

        csv_path = out_data_path / f"{source_file.stem}-{organelle_name}-zernike-stats.csv"
        z_stats.to_csv(csv_path)

        d_stats = get_depth_stats(        
                cellmask_obj,
                organelle_mask,
                organelle_obj,
                organelle_img,
                organelle_name,
                nuclei_obj
                )
        csv_path = out_data_path / f"{source_file.stem}-{organelle_name}-depth-stats.csv"
        d_stats.to_csv(csv_path)
        count += 1

    print(f"dumped {count}x3 projection stats csvs")
    return count


# infer_subc/organelles

from .nuclei import (
    infer_nuclei_fromlabel,
    fixed_infer_nuclei_fromlabel,
    infer_and_export_nuclei,
    infer_nuclei_fromlabel,
    get_nuclei,
)
from .cellmask import (
    infer_cellmask_fromcomposite,
    fixed_infer_cellmask_fromcomposite,
    non_linear_cellmask_transform,
    raw_cellmask_fromaggr,
    choose_max_label_cellmask_union_nucleus,
    infer_and_export_cellmask,
    get_cellmask,
)

# from .cellmask import infer_cellmask, fixed_infer_cellmask
from .cytoplasm import infer_cytoplasm, infer_and_export_cytoplasm, get_cytoplasm
from .lysosome import (
    infer_lyso,
    fixed_infer_lyso,
    lyso_filiment_filter,
    lyso_spot_filter,
    infer_and_export_lyso,
    get_lyso,
)
from .mitochondria import infer_mito, fixed_infer_mito, infer_and_export_mito, get_mito
from .golgi import infer_golgi, fixed_infer_golgi, infer_and_export_golgi, get_golgi
from .peroxisome import infer_perox, fixed_infer_perox, infer_and_export_perox, get_perox
from .er import (
    infer_ER,
    fixed_infer_ER,
    infer_and_export_ER,
    get_ER,
)
from .lipid import infer_LD, fixed_infer_LD, infer_and_export_LD, get_LD



import numpy as np
from skimage.filters import threshold_triangle, threshold_otsu, threshold_li, threshold_multiotsu, threshold_sauvola

# from skimage.filters import threshold_triangle, threshold_otsu, threshold_li, threshold_multiotsu, threshold_sauvola
from scipy.ndimage import median_filter, extrema, distance_transform_edt, sum, minimum_filter, maximum_filter

from aicssegmentation.core.utils import size_filter, hole_filling
from aicssegmentation.core.vessel import vesselness2D
from aicssegmentation.core.MO_threshold import MO

from aicssegmentation.core.vessel import filament_2d_wrapper, filament_3d_wrapper
from aicssegmentation.core.pre_processing_utils import (
    image_smoothing_gaussian_slice_by_slice,
    # edge_preserving_smoothing_3d,
)
from aicssegmentation.core.seg_dot import dot_2d_slice_by_slice_wrapper, dot_3d_wrapper

from typing import Tuple, List, Union, Any

from skimage.morphology import remove_small_objects, white_tophat, ball, disk, black_tophat, label
from skimage.segmentation import clear_border, watershed

from infer_subc.constants import (
    TEST_IMG_N,
    NUC_CH,
    LYSO_CH,
    MITO_CH,
    GOLGI_CH,
    PEROX_CH,
    ER_CH,
    LD_CH,
    RESIDUAL_CH,
)


def stack_layers(*layers) -> np.ndarray:
    """wrapper to stack the inferred objects into a single numpy.ndimage"""

    return np.stack(layers, axis=0)


def stack_masks(nuc_mask: np.ndarray, cellmask: np.ndarray, cyto_mask: np.ndarray) -> np.ndarray:
    """stack canonical masks:  cellmask, nuc, cytoplasm as uint8 (never more than 255 nuclei)"""
    layers = [nuc_mask, cellmask, cyto_mask]
    return np.stack(layers, axis=0).astype(np.uint8)


# TODO: check that the "noise" for the floor is correct... inverse_log should remove it?
def log_transform(image: np.ndarray) -> Tuple[np.ndarray, dict]:
    """Renormalize image intensities to log space

    Parameters
    ------------
    image:
        a 3d image

    Returns
    -------------
        Returns a tuple of transformed image and a dictionary to be passed into
        inverse_log_transform. The minimum and maximum from the dictionary
        can be applied to an image by the inverse_log_transform to
        convert it back to its former intensity values.
    """
    orig_min, orig_max = extrema(image)[:2]
    #
    # We add 1/2 bit noise to an 8 bit image to give the log a bottom
    #
    limage = image.copy()
    noise_min = orig_min + (orig_max - orig_min) / 256.0 + np.finfo(image.dtype).eps
    limage[limage < noise_min] = noise_min
    d = {"noise_min": noise_min}
    limage = np.log(limage)
    log_min, log_max = extrema(limage)[:2]
    d["log_min"] = log_min
    d["log_max"] = log_max
    return stretch(limage), d


def inverse_log_transform(image: np.ndarray, d: dict) -> np.ndarray:
    """Convert the values in image back to the scale prior to log_transform

    Parameters
    ------------
    image:
        a 3d image
    d:
        dictionary returned by log_transform

    Returns
    -------------
        de-logged image (np.ndarray)
    """
    return np.exp(unstretch(image, d["log_min"], d["log_max"]))


def stretch(image: np.ndarray, mask: Union[np.ndarray, None] = None) -> np.ndarray:
    """Normalize an image to make the minimum zero and maximum one

    Parameters
    ------------
    image:
        a 3d image to be normalized
    mask:
        optional mask of relevant pixels. None (default) means don't mask

    Returns
    -------------
        stretched (normalized to [0,1]) image (np.ndarray)

    """
    image = np.array(image, float)
    if np.product(image.shape) == 0:
        return image
    if mask is None:
        minval = np.min(image)
        maxval = np.max(image)
        if minval == maxval:
            if minval < 0:
                return np.zeros_like(image)
            elif minval > 1:
                return np.ones_like(image)
            return image
        else:
            return (image - minval) / (maxval - minval)
    else:
        significant_pixels = image[mask]
        if significant_pixels.size == 0:
            return image
        minval = np.min(significant_pixels)
        maxval = np.max(significant_pixels)
        if minval == maxval:
            transformed_image = minval
        else:
            transformed_image = (significant_pixels - minval) / (maxval - minval)
        result = image.copy()
        image[mask] = transformed_image
        return image


def unstretch(image: np.ndarray, minval: Union[int, float], maxval: Union[int, float]) -> np.ndarray:
    """Perform the inverse of stretch, given a stretched image

    Parameters
    ------------
    image:
        an image stretched by stretch or similarly scaled value or values
    minval:
        minimum of previously stretched image
    maxval:
        maximum of previously stretched image

    Returns
    -------------
        stretched (normalized to [0,1]) image (np.ndarray)
    """
    return image * (maxval - minval) + minval


def threshold_li_log(image_in: np.ndarray) -> np.ndarray:
    """
    thin wrapper to log-scale and inverse log image for threshold finding using li minimum cross-entropy

    Parameters
    ------------
    image:
        an np.ndarray
    Returns
    -------------
        boolean np.ndarray

    """
    image, d = log_transform(image_in.copy())
    threshold = threshold_li(image)
    threshold = inverse_log_transform(threshold, d)
    return threshold

    # image_out = log_transform(image_in.copy())
    # li_thresholded = structure_img_smooth > threshold_li_log(structure_img_smooth)


def threshold_otsu_log(image_in):
    """
    thin wrapper to log-scale and inverse log image for threshold finding using otsu

    Parameters
    ------------
    image:
        an np.ndarray
    Returns
    -------------
        boolean np.ndarray
    """
    image, d = log_transform(image_in.copy())
    threshold = threshold_otsu(image)
    threshold = inverse_log_transform(threshold, d)
    return threshold


def threshold_multiotsu_log(image_in):
    """
    thin wrapper to log-scale and inverse log image for threshold finding using otsu

    Parameters
    ------------
    image:
        an np.ndarray
    Returns
    -------------
        boolean np.ndarray
    """
    image, d = log_transform(image_in.copy())
    thresholds = threshold_multiotsu(image)
    thresholds = inverse_log_transform(thresholds, d)
    return thresholds


def masked_object_thresh(
    structure_img_smooth: np.ndarray, global_method: str, cutoff_size: int, local_adjust: float
) -> np.ndarray:
    """
    wrapper for applying Masked Object Thresholding with just two parameters via `MO` from `aicssegmentation`

    Parameters
    ------------
    structure_img_smooth:
        a 3d image
    global_method:
         which method to use for calculating global threshold. Options include:
         "triangle", "median", and "ave_tri_med".
         "ave_tri_med" refers the average of "triangle" threshold and "mean" threshold.
    cutoff_size:
        Masked Object threshold `size_min`
    local_adjust:
        Masked Object threshold `local_adjust`

    Returns
    -------------
        np.ndimage

    """
    struct_obj = MO(
        structure_img_smooth,
        object_minArea=cutoff_size,
        global_thresh_method=global_method,
        extra_criteria=True,
        local_adjust=local_adjust,
        return_object=False,
        dilate=False,  # WARNING: dilate=True causes a bug if there is only one Z
    )
    return struct_obj


def get_interior_labels(img_in: np.ndarray) -> np.ndarray:
    """
    gets the labeled objects from the X,Y "interior" of the image. We only want to clear the objects touching the sides of the volume, but not the top and bottom, so we pad and crop the volume along the 0th axis

    Parameters
    ------------
    img_in:
        a 3d image

    Returns
    -------------
        np.ndimage of labeled segmentations NOT touching the sides

    """
    segmented_padded = np.pad(
        label(img_in),
        ((1, 1), (0, 0), (0, 0)),
        mode="constant",
        constant_values=0,
    )
    interior = clear_border(segmented_padded)[1:-1]
    return interior
    # relabel?
    # return label(interior).astype(np.uint16)


def label_uint16(in_obj: np.ndarray) -> np.ndarray:
    """
    label segmentation and return as uint16

    Parameters
    ------------
    in_obj:
        a 3d image segmentaiton

    Returns
    -------------
        np.ndimage of labeled segmentations as np.uint16

    """
    if in_obj.dtype == "bool":
        return label(in_obj).astype(np.uint16)
    else:  # in_obj.dtype == np.uint8:
        return label(in_obj > 0).astype(np.uint16)


def label_bool_as_uint16(in_obj: np.ndarray) -> np.ndarray:
    """
    label segmentation and return as uint16

    Parameters
    ------------
    in_obj:
        a 3d image segmentaiton

    Returns
    -------------
        np.ndimage of labeled segmentations as np.uint16

    """
    return (in_obj > 0).astype(np.uint16)


def median_filter_slice_by_slice(struct_img: np.ndarray, size: int) -> np.ndarray:
    """
    wrapper for applying 2D median filter slice by slice on a 3D image

    Parameters
    ------------
    img:
        a 3d image

    size:
        the linear "size" which will be squared for

    Returns
    -------------
        np.ndimage

    """
    structure_img_denoise = np.zeros_like(struct_img)

    # this might be faster:  scipy.signal.medfilt2d()
    for zz in range(struct_img.shape[0]):
        structure_img_denoise[zz, :, :] = median_filter(struct_img[zz, :, :], size=size)

    return structure_img_denoise


def min_max_intensity_normalization(struct_img: np.ndarray) -> np.ndarray:
    """Normalize the intensity of input image so that the value range is from 0 to 1.

    Parameters
    ------------
    img:
        a 3d image

    Returns
    -------------
        np.ndimage
    """
    strech_min = struct_img.min()
    strech_max = struct_img.max()
    # do we need to convert to float?
    # #.astype(np.double)
    struct_img = (struct_img - strech_min + 1e-8) / (strech_max - strech_min + 1e-8)

    return struct_img


# # depriated. using media/gauss filtering for ER now
# def normalized_edge_preserving_smoothing(img_in: np.ndarray) -> np.ndarray:
#     """wrapper to min-max normalize + aicssegmentaion. edge_preserving_smoothing_3d

#     Parameters
#     ------------
#     img:
#         a 3d image

#     Returns
#     -------------
#         smoothed and normalized np.ndimage
#     """
#     struct_img = min_max_intensity_normalization(img_in)
#     # edge-preserving smoothing (Option 2, used for Sec61B)
#     struct_img = edge_preserving_smoothing_3d(struct_img)
#     # 10 seconds!!!!  tooooo slow... for maybe no good reason
#     return min_max_intensity_normalization(struct_img)


def weighted_aggregate(img_in: np.ndarray, *weights: int) -> np.ndarray:
    """
    helper to find weighted sum images
    Parameters
    ------------
    img_in:
        a 3d imagenp.ndarray
    *weights:
        list of integer weights to apply to our channels.  if the weights are less than 1, they will NOT be included (and might not be there)

    Returns
    -------------
        np.ndimage of weighted sum of channels

    """

    img_out = np.zeros_like(img_in[0]).astype(np.double)
    for ch, w in enumerate(weights):
        if w > 0:
            img_out += (w * 1.0) * img_in[ch]

    return img_out


def make_aggregate(
    img_in: np.ndarray,
    weight_ch0: int = 0,
    weight_ch1: int = 0,
    weight_ch2: int = 0,
    weight_ch3: int = 0,
    weight_ch4: int = 0,
    weight_ch5: int = 0,
    weight_ch6: int = 0,
    weight_ch7: int = 0,
    weight_ch8: int = 0,
    weight_ch9: int = 0,
    rescale: bool = True,
) -> np.ndarray:
    """define multi_channel aggregate.  weighted sum wrapper (plugin)

    Parameters
    ------------
    w0,w1,w2,w3,w4,w5,w6,w7,w8,w9
        channel weights
    rescale:
        scale to [0,1] if True. default True

    Returns
    -------------
        np.ndarray scaled aggregate

    """
    weights = (weight_ch0, weight_ch1, weight_ch2, weight_ch3, weight_ch4,
               weight_ch5, weight_ch6, weight_ch7, weight_ch8, weight_ch9)
    if rescale:
        # TODO: might NOT overflow here... maybe NOT do the normaization first?
        return min_max_intensity_normalization(weighted_aggregate(min_max_intensity_normalization(img_in), *weights))
    else:
        return weighted_aggregate(img_in, *weights)


def apply_threshold(
    img_in: np.ndarray,
    method: str = "otsu",
    thresh_factor: float = 1.0,
    thresh_min: Union[None, float] = None,
    thresh_max: Union[None, float] = None,
) -> np.ndarray:
    """return a binary mask after applying a log_li threshold

    Parameters
    ------------
    img_in:
        np.ndarray input image
    method:
        method for applying threshold.  "otsu"  or "li" (default), "triangle", "median", "ave", "sauvola","multi_otsu","muiltiotsu"
    thresh_factor:
        scaling value for threshold, defaults=1.0
    thresh_min
        absolute minumum for threshold, default=None
    thresh_max
        absolute maximum for threshold, default=None

    Returns
    -------------
        thresholded boolean np.ndarray
    """

    if method == "tri" or method == "triangle":
        threshold_val = threshold_triangle(img_in)
    elif method == "med" or method == "median":
        threshold_val = np.percentile(img_in, 50)
    elif method == "ave" or method == "ave_tri_med":
        global_tri = threshold_triangle(img_in)
        global_median = np.percentile(img_in, 50)
        threshold_val = (global_tri + global_median) / 2
    elif method == "li" or method == "cross_entropy" or method == "crossentropy":
        threshold_val = threshold_li(img_in)
    elif method == "sauvola":
        threshold_val = threshold_sauvola(img_in)
    elif method == "mult_otsu" or method == "multiotsu":
        threshold_val = threshold_multiotsu(img_in)
    else:  # default to "otsu"
        threshold_val = threshold_otsu(img_in)

    threshold = threshold_val * thresh_factor

    if thresh_min is not None:
        threshold = max(threshold, thresh_min)
    if thresh_max is not None:
        threshold = min(threshold, thresh_max)
    return img_in > threshold


def apply_log_li_threshold(
    img_in: np.ndarray,
    thresh_factor: float = 1.0,
    thresh_min: Union[None, float] = None,
    thresh_max: Union[None, float] = None,
) -> np.ndarray:
    """return a binary mask after applying a log_li threshold

    Parameters
    ------------
    img_in:
        input ndimage array (np.ndimage)
    thresh_factor:
        scaling value for threshold, defaults=1.0
    thresh_min
        absolute minumum for threshold, default=None
    thresh_max
        absolute maximum for threshold, default=None

    Returns
    -------------
        thresholded boolean np.ndarray
    """
    # struct_obj = struct_img > filters.threshold_li(struct_img)
    threshold_value_log = threshold_li_log(img_in)
    threshold = threshold_value_log * thresh_factor

    if thresh_min is not None:
        threshold = max(threshold, thresh_min)
    if thresh_max is not None:
        threshold = min(threshold, thresh_max)
    return img_in > threshold


# NOTE this is identical to veselnessSliceBySlice from aicssegmentation.core.vessel
def vesselness_slice_by_slice(nd_array: np.ndarray, sigma: float, cutoff: float = -1, tau: float = 0.75):
    """
    wrapper for applying multi-scale 2D filament filter on 3D images in a
    slice by slice fashion,  Note that it only performs at a single scale....     NOTE: The paramater
    whiteonblack = True is hardcoded which sets the filamentous structures are bright on dark background

    Parameters
    -----------
    nd_array:
        the 3D image to be filterd on
    sigma:
        single scale to use
    cutoff:
        the cutoff value to apply on the filter result. If the cutoff is
        negative, no cutoff will be applied. Default is -1.
    tau:
        parameter that controls response uniformity. The value has to be
        between 0.5 and 1. Lower tau means more intense output response.
        Default is 0.5
    """

    # # this hack is to accomodate the workflow widgets
    # if not isinstance(sigmas, List):
    #     sigmas = [sigmas]

    mip = np.amax(nd_array, axis=0)
    response = np.zeros(nd_array.shape)
    for zz in range(nd_array.shape[0]):
        tmp = np.concatenate((nd_array[zz, :, :], mip), axis=1)
        tmp = vesselness2D(tmp, sigmas=[sigma], tau=tau, whiteonblack=True)
        response[zz, :, : nd_array.shape[2] - 3] = tmp[:, : nd_array.shape[2] - 3]

    if cutoff < 0:
        return response
    else:
        return response > cutoff


def select_channel_from_raw(img_in: np.ndarray, chan: Union[int, Tuple[int]]) -> np.ndarray:
    """ "
    select channel from multi-channel 3D image (np.ndarray)
    Parameters
    ------------
    img_in :
        the 3D image to be filterd on
    chan :
        channel to extract.

    Returns
    -------------
        np.ndarray
    """
    return img_in[chan]


def select_z_from_raw(img_in: np.ndarray, z_slice: Union[int, Tuple[int]]) -> np.ndarray:
    """
    select Z-slice from 3D multi-channel image (np.ndarray)

    Parameters
    ------------
    img_in :
        the 3D image to be filterd on
    chan :
        channel to extract.

    Returns
    -------------
        np.ndarray
    """
    if isinstance(z_slice, int):
        z_slice = [z_slice]
    else:
        z_slice = list(z_slice)

    return img_in[:, z_slice, :, :]


def scale_and_smooth(
    img_in: np.ndarray, median_size: int = 1, gauss_sigma: float = 1.34, slice_by_slice: bool = True
) -> np.ndarray:
    """
    helper to perform min-max scaling, and median+gaussian smoothign all at once
    Parameters
    ------------
    img_in: np.ndarray
        a 3d image
    median_size: int
        width of median filter for signal
    gauss_sigma: float
        sigma for gaussian smoothing of  signal
    slice_by_slice:
        NOT IMPLIMENTED.  toggles whether to do 3D operations or slice by slice in Z

    Returns
    -------------
        np.ndimage

    """
    img = min_max_intensity_normalization(img_in.copy())  # is this copy nescesa

    # TODO:  make non-slice-by-slice work
    slice_by_slice = True
    if slice_by_slice:
        if median_size > 1:
            img = median_filter_slice_by_slice(img, size=median_size)
        img = image_smoothing_gaussian_slice_by_slice(img, sigma=gauss_sigma)
    else:
        print(" PLEASE CHOOOSE 'slice-by-slice', 3D is not yet implimented")

    return img


# DEPRICATED
def aggregate_signal_channels(
    img_in: np.ndarray, chs: Union[List, Tuple], ws: Union[List, Tuple, Any] = None
) -> np.ndarray:
    """
    return a weighted sum of the image across channels (DEPRICATED)

    Parameters
    ------------
    img_in:
        np.ndarray  [ch,z,x,y]
    chs:
        list/tuple of channels to aggregate
    ws:
        list/tuple/ of weights for aggregation

    Returns
    -------------
        np.ndarray
    """
    n_chan = len(chs)
    if n_chan <= 1:
        return img_in[chs]

    if ws is None:
        ws = n_chan * [1.0]
    img_out = np.zeros_like(img_in[0]).astype(np.double)
    for w, ch in zip(ws, chs):
        img_out += w * img_in[ch]
    return img_out
    # return img_in[ch_to_agg,].astype( np.double ).sum(axis=0)


def choose_agg_signal_zmax(img_in: np.ndarray, chs: List[int], ws=None, mask=None) -> np.ndarray:
    """
    return z the maximum signal for the aggregate signal

    Parameters
    ------------
    img_in:
        np.ndarray  [ch,z,x,y]
    chs:
        list of channels to aggregate
    ws:
        list of weights for aggregation
    mask:
        mask for img_in

    Returns
    -------------
        np.ndarray z with maximum signal
    """
    total_florescence_ = aggregate_signal_channels(img_in, chs)
    if mask is not None:
        total_florescence_[mask] = 0.0
    return int(total_florescence_.sum(axis=(1, 2)).argmax())


def masked_inverted_watershed(img_in: np.ndarray, 
                                     markers: np.ndarray, 
                                     mask: np.ndarray, 
                                     method: str='slice-by-slice'):
    """wrapper for watershed on inverted image and masked

    Parameters
    ------------
    in_img:
        a 3d image
    markers: 
        objects used to seed the watershed
    mask:
        instance segmentation of the in_img
    method:
        'slice-by-slice' results in a connectivity of np.ones((1,3,3), bool); '3D' results in a connectivity of np.ones((3,3,3), bool)

    """
    if method == 'slice-by-slice':    
        labels_out = watershed(
            1.0 - img_in,
            markers=markers,
            connectivity=np.ones((1, 3, 3), bool),
            mask=mask)
    elif method == '3D':
        labels_out = watershed(
        1.0 - img_in,
        markers=markers,
        connectivity=np.ones((3, 3, 3), bool),
        mask=mask)
    else:
        print(f"incompatable method: {method}")

    return labels_out



def choose_max_label(
    raw_signal: np.ndarray, labels_in: np.ndarray, target_labels: Union[np.ndarray, None] = None
) -> np.ndarray:
    """
    keep only the segmentation corresponding to the maximum raw signal.  candidate  label is taken from target_labels if not None

    Parameters
    ------------
    raw_signal:
        the image to filter on
    labels_in:
        segmentation labels
    target_labels:
        labels to consider

    Returns
    -------------
        np.ndarray of labels corresponding to the largest total signal

    """
    keep_label = get_max_label(raw_signal, labels_in, target_labels)
    labels_max = np.zeros_like(labels_in)
    labels_max[labels_in == keep_label] = 1
    return labels_max


def get_max_label(
    raw_signal: np.ndarray, labels_in: np.ndarray, target_labels: Union[np.ndarray, None] = None
) -> np.ndarray:
    """
    keep only the label with the maximum raw signal.  candidate  label is taken from target_labels if not None

    Parameters
    ------------
    raw_signal:
        the image to filter on
    labels_in:
        segmentation labels
    target_labels:
        labels to consider

    Returns
    -------------
        np.ndarray of labels corresponding to the largest total signal

    """
    if target_labels is None:
        all_labels = np.unique(labels_in)[1:]
    else:
        all_labels = np.unique(target_labels)[1:]

    total_signal = [raw_signal[labels_in == label].sum() for label in all_labels]
    # combine NU and "labels" to make a CELLMASK
    keep_label = all_labels[np.argmax(total_signal)]

    return keep_label


def fill_and_filter_linear_size(
    img: np.ndarray, hole_min: int, hole_max: int, min_size: int, method: str = "slice_by_slice", connectivity: int = 1
) -> np.ndarray:
    """wraper to aiscsegmentation `hole_filling` and `size_filter` with size argument in linear units

    Parameters
    ------------
    img:
        the image to filter on
    hole_min: int
        the minimum width of the holes to be filled
    hole_max: int
        the maximum width of the holes to be filled
    min_size: int
        the minimum size expressed as 1D length (so squared for slice-by-slice, cubed for 3D)
    method: str
        either "3D" or "slice_by_slice", default is "slice_by_slice"
    connnectivity: int
        the connectivity to use when computing object size
    Returns
    -------------
        a binary image after hole filling and filtering small objects; np.ndarray
    """
    if not img.any():
        return img

    if method == "3D":
        if hole_max > 0:
            img = hole_filling(img, hole_min=hole_min**3, hole_max=hole_max**3, fill_2d=False)
        return size_filter(img, min_size=min_size**3, method="3D", connectivity=connectivity)
    elif method == "slice_by_slice":
        if hole_max > 0:
            img = hole_filling(img, hole_min=hole_min**2, hole_max=hole_max**2, fill_2d=True)
        return size_filter(img, min_size=min_size**2, method="slice_by_slice", connectivity=connectivity)
    else:
        print(f"undefined method: {method}")


def size_filter_linear_size(
    img: np.ndarray, min_size: int, method: str = "slice_by_slice", connectivity: int = 1
) -> np.ndarray:
    """size filter wraper to aiscsegmentation `size_filter` with size argument in linear units

    Parameters
    ------------
    img:
        the image to filter on
    min_size: int
        the minimum size expressed as 1D length (so squared for slice-by-slice, cubed for 3D)
    method: str
        either "3D" or "slice_by_slice", default is "slice_by_slice"
    connnectivity: int
        the connectivity to use when computing object size
    Returns
    -------------
        np.ndarray
    """
    # return remove_small_objects(img > 0, min_size=min_size, connectivity=connectivity, in_place=False)
    if not img.any():
        return img

    if method == "3D":
        return size_filter(img, min_size=min_size**3, method="3D", connectivity=connectivity)
    elif method == "slice_by_slice":
        return size_filter(img, min_size=min_size**2, method="slice_by_slice", connectivity=connectivity)
    else:
        raise NotImplementedError(f"unsupported method {method}")


def hole_filling_linear_size(img: np.ndarray, hole_min: int, hole_max: int, fill_2d=True) -> np.ndarray:
    """Fill holes  wraper to aiscsegmentation `hole_filling` with size argument in linear units.  always does slice-by-slice

    Parameters
    ------------
    img:
        the image to filter on
    hole_min: int
        the minimum width of the holes to be filled
    hole_max: int
        the maximum width of the holes to be filled

    Returns
    -----------
        a binary image after hole filling
    """
    if fill_2d:
        return hole_filling(img, hole_min=hole_min**2, hole_max=hole_max**2, fill_2d=True)
    else:
        return hole_filling(img, hole_min=hole_min**3, hole_max=hole_max**3, fill_2d=False)


def apply_mask(img_in: np.ndarray, mask: np.ndarray) -> np.ndarray:
    """mask the image

    Parameters
    ------------
    img_in:
        the image to filter on
    mask:
        the mask to apply

    Returns
    -----------
    img_out:
        a new (copied) array with mask applied
    """
    assert img_in.shape == mask.shape

    img_out = img_in.copy()
    if mask.dtype == "bool":
        img_out[~mask] = 0
    else:
        img_out[mask < 1] = 0

    return img_out


def enhance_speckles(image: np.ndarray, radius: int, volumetric: bool = False) -> np.ndarray:
    """enhance "spreckles" small dots

    Parameters
    ------------
    image: np.ndarray
        the image to filter on
    radius: int
        radius of the "filter"
    volumetric: bool
        True for 3D analysis

    Returns
    -----------
    result:
        filtered boolean np.ndarray
    """
    radius = radius / 2
    if volumetric:
        selem = ball(radius)
    else:
        selem = disk(radius)

    # if radius >10:
    #         minimum = scipy.ndimage.minimum_filter(image, footprint=selem)
    #         maximum = scipy.ndimage.maximum_filter(minimum, footprint=selem)
    #         result = data - maximum
    # else:
    result = white_tophat(image)

    return result


def enhance_neurites(image: np.ndarray, radius: int, volumetric: bool = False) -> np.ndarray:
    """enhance "neurites" or filiments

    Parameters
     ------------
     image: np.ndarray
         the image to filter on
     radius: int
         radius of the "filter"
     volumetric: bool
         True for 3D analysis

     Returns
     -----------
     result:
         filtered boolean np.ndarray
    """
    if volumetric:
        selem = ball(radius)
    else:
        selem = disk(radius)
    white = white_tophat(image, selem)
    black = black_tophat(image, selem)
    result = image + white - black
    result[result > 1] = 1
    result[result < 0] = 0

    return result


def filament_filter_3(in_img: np.ndarray, 
                       filament_scale_1: float, 
                       filament_cutoff_1: float,
                       filament_scale_2: float, 
                       filament_cutoff_2: float,
                       filament_scale_3: float, 
                       filament_cutoff_3: float,
                       method: str
                       ) -> np.ndarray:
    """filament filter helper function for 3 levels (scale+cut). filter pairs are run if scale is > 0.

    Parameters
    ------------
    in_img:
        the image to filter on np.ndarray
    filament_scale_1:
        scale or size of the "filter" float
    filament_cutoff_1:
        cutoff for thresholding float
    filament_scale_2:
        scale or size of the "filter" float
    filament_cutoff_2:
        cutoff for thresholding float
    filament_scale_3:
        scale or size of the "filter" float
    filament_cutoff_3:
        cutoff for thresholding float
    method:
        either "3D" or "slice_by_slice", default is "slice_by_slice"

    Returns
    -----------
    result:
        filtered boolean np.ndarray

    """
    scales = [filament_scale_1, filament_scale_2, filament_scale_3]
    cuts = [filament_cutoff_1, filament_cutoff_2, filament_cutoff_3]
    f_param = [[sc, ct] for sc, ct in zip(scales, cuts) if sc > 0]

    if method == "3D":
        seg = filament_3d_wrapper(in_img, f_param)
    elif method == "slice_by_slice":
        seg = filament_2d_wrapper(in_img, f_param)
    else:
        print(f"undefined method: {method}")

    return seg


def filament_filter(in_img: np.ndarray, filament_scale: float, filament_cut: float) -> np.ndarray:
    """filament wrapper to properly pack parameters into filament_2d_wrapper

    Parameters
    ------------
    in_img:
        the image to filter on np.ndarray
    filament_scale:
        scale or size of the "filter" float
    filament_cut:
        cutoff for thresholding float

    Returns
    -----------
    result:
        filtered boolean np.ndarray

    """
    f2_param = [[filament_scale, filament_cut]]
    # f2_param = [[1, 0.15]]  # [scale_1, cutoff_1]
    return filament_2d_wrapper(in_img, f2_param)


# def dot_filter_3(
#     in_img: np.ndarray,
#     dot_scale_1: float,
#     dot_cut_1: float,
#     dot_scale_2: float,
#     dot_cut_2: float,
#     dot_scale_3: float,
#     dot_cut_3: float,
# ) -> np.ndarray:
#     """spot filter helper function for 3 levels (scale+cut).  if scale_i is > 0.0001 its skipped

#     Parameters
#     ------------
#     in_img:
#         a 3d  np.ndarray image of the inferred organelle (labels or boolean)
#     dot_scale_1:
#         scale or size of the "filter" float
#     dot_cut_1:
#         cutoff for thresholding float
#     dot_scale_2:
#         scale or size of the "filter" float
#     dot_cut_2:
#         cutoff for thresholding float
#     dot_scale_3:
#         scale or size of the "filter" float
#     dot_cut_3:
#         cutoff for thresholding float

#     Returns
#     -------------
#     segmented dots over 3 scales

#     """
#     scales = [dot_scale_1, dot_scale_2, dot_scale_3]
#     cuts = [dot_cut_1, dot_cut_2, dot_cut_3]

#     s2_param = [[sc, ct] for sc, ct in zip(scales, cuts) if sc > 0.0001]
#     # s2_param = [[dot_scale1, dot_cut1], [dot_scale2, dot_cut2], [dot_scale3, dot_cut3]]
#     return dot_2d_slice_by_slice_wrapper(in_img, s2_param)


def dot_filter_3(
    in_img: np.ndarray,
    dot_scale_1: float,
    dot_cutoff_1: float,
    dot_scale_2: float,
    dot_cutoff_2: float,
    dot_scale_3: float,
    dot_cutoff_3: float,
    method: str = "slice_by_slice"
) -> np.ndarray:
    """spot filter helper function for 3 levels (scale+cut). filter pairs are run if scale is > 0.

    Parameters
    ------------
    in_img:
        a 3d  np.ndarray image of the inferred organelle (labels or boolean)
    dot_scale_1:
        scale or size of the "filter" float
    dot_cutoff_1:
        cutoff for thresholding float
    dot_scale_2:
        scale or size of the "filter" float
    dot_cutoff_2:
        cutoff for thresholding float
    dot_scale_3:
        scale or size of the "filter" float
    dot_cut_3:
        cutoff for thresholding float
    method:
        either "3D" or "slice_by_slice", default is "slice_by_slice"

    Returns
    -------------
    segmented dots over 3 scales

    """
    scales = [dot_scale_1, dot_scale_2, dot_scale_3]
    cuts = [dot_cutoff_1, dot_cutoff_2, dot_cutoff_3]
    s_param = [[sc, ct] for sc, ct in zip(scales, cuts) if sc > 0]

    if method == "3D":
        seg = dot_3d_wrapper(in_img, s_param)
    elif method == "slice_by_slice":
        seg = dot_2d_slice_by_slice_wrapper(in_img, s_param)
    else:
        print(f"undefined method: {method}")

    return seg

# centrosome routines


def size_similarly(labels, secondary):
    """Size the secondary matrix similarly to the labels matrix

    labels - labels matrix
    secondary - a secondary image or labels matrix which might be of
                different size.
    Return the resized secondary matrix and a mask indicating what portion
    of the secondary matrix is bogus (manufactured values).

    Either the mask is all ones or the result is a copy, so you can
    modify the output within the unmasked region w/o destroying the original.
    """
    if labels.shape[:2] == secondary.shape[:2]:
        return secondary, np.ones(secondary.shape, bool)
    if labels.shape[0] <= secondary.shape[0] and labels.shape[1] <= secondary.shape[1]:
        if secondary.ndim == 2:
            return (
                secondary[: labels.shape[0], : labels.shape[1]],
                np.ones(labels.shape, bool),
            )
        else:
            return (
                secondary[: labels.shape[0], : labels.shape[1], :],
                np.ones(labels.shape, bool),
            )

    #
    # Some portion of the secondary matrix does not cover the labels
    #
    result = np.zeros(list(labels.shape) + list(secondary.shape[2:]), secondary.dtype)
    i_max = min(secondary.shape[0], labels.shape[0])
    j_max = min(secondary.shape[1], labels.shape[1])
    if secondary.ndim == 2:
        result[:i_max, :j_max] = secondary[:i_max, :j_max]
    else:
        result[:i_max, :j_max, :] = secondary[:i_max, :j_max, :]
    mask = np.zeros(labels.shape, bool)
    mask[:i_max, :j_max] = 1
    return result, mask


def distance_to_edge(labels):
    """Compute the distance of a pixel to the edge of its object

    labels - a labels matrix

    returns a matrix of distances
    """
    colors = color_labels(labels)
    max_color = np.max(colors)
    result = np.zeros(labels.shape)
    if max_color == 0:
        return result

    for i in range(1, max_color + 1):
        mask = colors == i
        result[mask] = distance_transform_edt(mask)[mask]
    return result


def color_labels(labels, distance_transform=False):
    """Color a labels matrix so that no adjacent labels have the same color

    distance_transform - if true, distance transform the labels to find out
         which objects are closest to each other.

    Create a label coloring matrix which assigns a color (1-n) to each pixel
    in the labels matrix such that all pixels similarly labeled are similarly
    colored and so that no similiarly colored, 8-connected pixels have
    different labels.

    You can use this function to partition the labels matrix into groups
    of objects that are not touching; you can then operate on masks
    and be assured that the pixels from one object won't interfere with
    pixels in another.

    returns the color matrix
    """
    if distance_transform:
        i, j = distance_transform_edt(labels == 0, return_distances=False, return_indices=True)
        dt_labels = labels[i, j]
    else:
        dt_labels = labels
    # Get the neighbors for each object
    v_count, v_index, v_neighbor = find_neighbors(dt_labels)
    # Quickly get rid of labels with no neighbors. Greedily assign
    # all of these a color of 1
    v_color = np.zeros(len(v_count) + 1, int)  # the color per object - zero is uncolored
    zero_count = v_count == 0
    if np.all(zero_count):
        # can assign all objects the same color
        return (labels != 0).astype(int)
    v_color[1:][zero_count] = 1
    v_count = v_count[~zero_count]
    v_index = v_index[~zero_count]
    v_label = np.argwhere(~zero_count).transpose()[0] + 1
    # If you process the most connected labels first and use a greedy
    # algorithm to preferentially assign a label to an existing color,
    # you'll get a coloring that uses 1+max(connections) at most.
    #
    # Welsh, "An upper bound for the chromatic number of a graph and
    # its application to timetabling problems", The Computer Journal, 10(1)
    # p 85 (1967)
    #
    sort_order = np.lexsort([-v_count])
    v_count = v_count[sort_order]
    v_index = v_index[sort_order]
    v_label = v_label[sort_order]
    for i in range(len(v_count)):
        neighbors = v_neighbor[v_index[i] : v_index[i] + v_count[i]]
        colors = np.unique(v_color[neighbors])
        if colors[0] == 0:
            if len(colors) == 1:
                # only one color and it's zero. All neighbors are unlabeled
                v_color[v_label[i]] = 1
                continue
            else:
                colors = colors[1:]
        # The colors of neighbors will be ordered, so there are two cases:
        # * all colors up to X appear - colors == np.arange(1,len(colors)+1)
        # * some color is missing - the color after the first missing will
        #   be mislabeled: colors[i] != np.arange(1, len(colors)+1)
        crange = np.arange(1, len(colors) + 1)
        misses = crange[colors != crange]
        if len(misses):
            color = misses[0]
        else:
            color = len(colors) + 1
        v_color[v_label[i]] = color
    return v_color[labels]


def find_neighbors(labels):
    """Find the set of objects that touch each object in a labels matrix

    Construct a "list", per-object, of the objects 8-connected adjacent
    to that object.
    Returns three 1-d arrays:
    * array of #'s of neighbors per object
    * array of indexes per object to that object's list of neighbors
    * array holding the neighbors.

    For instance, say 1 touches 2 and 3 and nobody touches 4. The arrays are:
    [ 2, 1, 1, 0], [ 0, 2, 3, 4], [ 2, 3, 1, 1]
    """
    max_label = np.max(labels)
    # Make a labels matrix with zeros around the edges so we can do index
    # offsets without worrying.
    #
    new_labels = np.zeros(np.array(labels.shape) + 2, labels.dtype)
    new_labels[1:-1, 1:-1] = labels
    labels = new_labels
    # Only consider the points that are next to others
    adjacent_mask = adjacent(labels)
    adjacent_i, adjacent_j = np.argwhere(adjacent_mask).transpose()
    # Get matching vectors of labels and neighbor labels for the 8
    # compass directions.
    count = len(adjacent_i)
    if count == 0:
        return (np.zeros(max_label, int), np.zeros(max_label, int), np.zeros(0, int))
    # The following bizarre construct does the following:
    # labels[adjacent_i, adjacent_j] looks up the label for each pixel
    # [...]*8 creates a list of 8 references to it
    # np.hstack concatenates, giving 8 repeats of the list
    v_label = np.hstack([labels[adjacent_i, adjacent_j]] * 8)
    v_neighbor = np.zeros(count * 8, int)
    index = 0
    for i, j in ((-1, -1), (-1, 0), (-1, 1), (0, -1), (0, 1), (1, -1), (1, 0), (1, 1)):
        v_neighbor[index : index + count] = labels[adjacent_i + i, adjacent_j + j]
        index += count
    #
    # sort by label and neighbor
    #
    sort_order = np.lexsort((v_neighbor, v_label))
    v_label = v_label[sort_order]
    v_neighbor = v_neighbor[sort_order]
    #
    # eliminate duplicates by comparing each element after the first one
    # to its previous
    #
    first_occurrence = np.ones(len(v_label), bool)
    first_occurrence[1:] = (v_label[1:] != v_label[:-1]) | (v_neighbor[1:] != v_neighbor[:-1])
    v_label = v_label[first_occurrence]
    v_neighbor = v_neighbor[first_occurrence]
    #
    # eliminate neighbor = self and neighbor = background
    #
    to_remove = (v_label == v_neighbor) | (v_neighbor == 0)
    v_label = v_label[~to_remove]
    v_neighbor = v_neighbor[~to_remove]
    #
    # The count of # of neighbors
    #
    v_count = fixup_scipy_ndimage_result(sum(np.ones(v_label.shape), v_label, np.arange(max_label, dtype=np.int32) + 1))
    v_count = v_count.astype(int)
    #
    # The index into v_neighbor
    #
    v_index = np.cumsum(v_count)
    v_index[1:] = v_index[:-1]
    v_index[0] = 0
    return (v_count, v_index, v_neighbor)


# from CellProfiler
def fixup_scipy_ndimage_result(whatever_it_returned):
    """Convert a result from scipy.ndimage to a numpy array

    scipy.ndimage has the annoying habit of returning a single, bare
    value instead of an array if the indexes passed in are of length 1.
    For instance:
    scipy.ndimage.maximum(image, labels, [1]) returns a float
    but
    scipy.ndimage.maximum(image, labels, [1,2]) returns a list
    """
    if getattr(whatever_it_returned, "__getitem__", False):
        return np.array(whatever_it_returned)
    else:
        return np.array([whatever_it_returned])


def adjacent(labels):
    """Return a binary mask of all pixels which are adjacent to a pixel of
    a different label.

    """
    high = labels.max() + 1
    if high > np.iinfo(labels.dtype).max:
        labels = labels.astype(np.int32)
    image_with_high_background = labels.copy()
    image_with_high_background[labels == 0] = high
    min_label = minimum_filter(
        image_with_high_background,
        footprint=np.ones((3, 3), bool),
        mode="constant",
        cval=high,
    )
    max_label = maximum_filter(labels, footprint=np.ones((3, 3), bool), mode="constant", cval=0)
    return (min_label != max_label) & (labels > 0)


def img_to_uint8(data_in: np.ndarray) -> np.ndarray:
    """
    helper to convert bask to `binary` uint8 (true -> 255) to accomodate napari default scaling
    """
    print(f"changing from {data_in.dtype} to np.uint8")
    data_in = data_in.astype(np.uint8)
    data_in[data_in > 0] = 1
    return data_in


def img_to_bool(data_in: np.ndarray) -> np.ndarray:
    """
    helper to make sure we are keeping track of things correctly
    """
    print(f"changing from {data_in.dtype} to bool")
    data_out = data_in > 0
    print(f"    -> {data_out.dtype}")
    return data_out


# # from cellprofiler / centrosome / smooth.py
# def smooth_with_function_and_mask(image, function, mask):
#     """Smooth an image with a linear function, ignoring the contribution of masked pixels

#     image - image to smooth
#     function - a function that takes an image and returns a smoothed image
#     mask  - mask with 1's for significant pixels, 0 for masked pixels

#     This function calculates the fractional contribution of masked pixels
#     by applying the function to the mask (which gets you the fraction of
#     the pixel data that's due to significant points). We then mask the image
#     and apply the function. The resulting values will be lower by the bleed-over
#     fraction, so you can recalibrate by dividing by the function on the mask
#     to recover the effect of smoothing from just the significant pixels.
#     """

#     not_mask = np.logical_not(mask)
#     bleed_over = function(mask.astype(float))
#     masked_image = np.zeros(image.shape, image.dtype)
#     masked_image[mask] = image[mask]
#     smoothed_image = function(masked_image)
#     output_image = smoothed_image / (bleed_over + np.finfo(float).eps)
#     return output_image

#     # blurred_image = centrosome.smooth.smooth_with_function_and_mask(
#     #     data,
#     #     lambda x: scipy.ndimage.gaussian_filter(x, sigma, mode="constant", cval=0),
#     #     mask,
#     # )



import numpy as np
import logging

from typing import Any, Dict, List
from infer_subc.exceptions import ArgumentNullError
from infer_subc.workflow.workflow_step import WorkflowStep
from infer_subc.workflow.workflow_definition import WorkflowDefinition
from napari.layers import Image

log = logging.getLogger(__name__)


class Workflow:
    """
    Represents an executable aics-segmentation workflow
    This class provides the functionality to run a workflow using an image input
    according to the steps defined in its WorkflowDefinition.
    """

    def __init__(self, workflow_definition: WorkflowDefinition, input_image: np.ndarray):
        if workflow_definition is None:
            raise ArgumentNullError("workflow_definition")
        if input_image is None:
            raise ArgumentNullError("input_image")
        self._definition: WorkflowDefinition = workflow_definition
        self._starting_image: np.ndarray = input_image
        self._next_step: int = 0  # Next step to execute
        self._results: List = list()  # Store most recent step results

    @property
    def workflow_definition(self) -> WorkflowDefinition:
        return self._definition

    def reset(self):
        """
        Reset the workflow so it can be run again
        """
        self._next_step = 0
        self._results = list()

    def get_next_step(self) -> WorkflowStep:
        """
        Get the next step to be performed

        Params:
            none: None

        Returns
            (WorkflowStep): next WorkflowStep object to perform on image
            None if all steps have already been executed
        """
        if self._next_step >= len(self._definition.steps):
            return None
        return self._definition.steps[self._next_step]

    def execute_next(self, parameters: Dict[str, Any] = None) -> np.ndarray:
        """
        Execute the next workflow step.

        Params:
            parameters: Optional dictionary of parameter inputs to use when executing the step
                        If parameters are not provided, the step's default parameters will be used

        Returns
            result (np.ndarray): resultant image from running the
            next workflow step
        """

        step = self.get_next_step()

        log.info(f"Executing step #{step.step_number}")
        # Pick which image to perform the workflow step on
        image: np.ndarray = None

        if self._next_step == 0:
            # First image, so use the starting image for the next workflow step
            image = [self._starting_image]
        elif self.is_done():
            # No more workflow steps to perform
            # TODO: what to do if done with workflow
            #  but execute_next is prompted?
            # printing message for now
            log.info("No steps left to run")
        else:
            image = list()
            for i in step.parent:
                res = self.get_result(i - 1)  # parents are 1 indexed
                image.append(res)

        result: np.ndarray = self.get_next_step().execute(image, parameters or step.parameter_values)
        self._results.append(result)

        # Only increment after running step
        self._next_step += 1
        return result

    # TODO maybe change this to match the step number instead?
    #      Review when we implement rerunning single workflow steps
    def get_result(self, step_index: int) -> np.ndarray:
        """
        Get the result image for a workflow step.

        You must call execute() on the workflow step in order to
        produce a result first before calling this function.

        Params:
            step_index (int): index of the WorkflowStep in the
            workflowengine to get the result image of.

        Returns
            self.image (np.ndarray): Result of performing workflow step
                                     on the given image
                                     None if step has not been executed yet.
        """
        if step_index < 0:
            return self._starting_image
        if step_index >= len(self._results):
            return None  # returns None if the WorkflowStep has not been executed.

        return self._results[step_index]

    def get_most_recent_result(self) -> np.ndarray:
        """
        Get the result from the last executed WorkflowStep.

        Params:
           none: None

        Returns
            (np.ndarray): Result of the last executed WorkflowStep,
                            returns the starting image if no Workflowsteps have
                            been run.
        """
        if self._next_step == 0:
            return self._starting_image  # TODO does this behavior make sense? Return None instead?
        else:
            return self.get_result(self._next_step - 1)

    def execute_all(self) -> np.ndarray:
        """
        Execute all steps in the Workflow
        Note: default parameters will be used to execute the steps. To execute a step
              with user-provided parameters, use execute_next()

        Params:
            none: None

        Returns
            (np.ndarray): Result of the final WorkflowStep.
        """
        self.reset()
        while not self.is_done():
            self.execute_next()
        return self.get_most_recent_result()

    def execute_step(self, i: int, parameters: Dict[str, Any], selected_image: List[Image]) -> np.ndarray:
        """

        Args:
            i: step number (0 indexed) that you want to run

        Returns

        """
        # TODO: discuss using selected layer for all types
        step_to_run = self._definition.steps[i]

        image = [i.data for i in selected_image]
        result: np.ndarray = step_to_run.execute(image, parameters or step_to_run.parameter_values)

        if len(self._results) <= i:
            # this is the first time running this step
            self._results.append(result)
        else:
            # this step has been run before so replacing old value
            self._results[i] = result

        return result

    def is_done(self) -> bool:
        """
        Check if all WorkflowSteps have been executed.

        Params:
            none: None

        Returns
            (bool): True if all WorkflowSteps have been executed, False if not
        """
        return self._next_step >= len(self._definition.steps)



import numpy as np
from typing import Tuple

from aicssegmentation.core.pre_processing_utils import image_smoothing_gaussian_slice_by_slice


from infer_subc.constants import (
    TEST_IMG_N,
    NUC_CH,
    LYSO_CH,
    MITO_CH,
    GOLGI_CH,
    PEROX_CH,
    ER_CH,
    LD_CH,
    RESIDUAL_CH,
)

from infer_subc.core.img import (
    median_filter_slice_by_slice,
    min_max_intensity_normalization,
    apply_log_li_threshold,
    choose_agg_signal_zmax,
    select_z_from_raw,
)


def get_optimal_Z_image(img_data: np.ndarray, nuc_ch: int, ch_to_agg: Tuple[int]) -> np.ndarray:
    """
    Procedure to infer _best_ Zslice from linearly unmixed input

    Parameters
    ------------
    in_img:
        a 3d image containing all the channels
    nuc_ch:
        channel with nuclei signal
    ch_to_agg:
        tuple of channels to aggregate for selecting Z

    Returns
    -------------
    image np.ndarray with single selected Z-slice   (Channels, 1, X, Y)

    """
    optimal_Z = find_optimal_Z(img_data, nuc_ch, ch_to_agg)
    return select_z_from_raw(img_data, optimal_Z)


def fixed_get_optimal_Z_image(img_data: np.ndarray) -> np.ndarray:
    """
    Procedure to infer _best_ Zslice from linearly unmixed input with fixed parameters
    """
    optimal_Z = fixed_find_optimal_Z(img_data)
    return select_z_from_raw(img_data, optimal_Z)


def fixed_find_optimal_Z(img_data: np.ndarray) -> int:
    """
    Procedure to infer _best_ Zslice from linearly unmixed input with fixed parameters
    """
    nuc_ch = NUC_CH
    ch_to_agg = (LYSO_CH, MITO_CH, GOLGI_CH, PEROX_CH, ER_CH, LD_CH)
    return find_optimal_Z(img_data, nuc_ch, ch_to_agg)


def find_optimal_Z(raw_img: np.ndarray, nuc_ch: int, ch_to_agg: Tuple[int]) -> int:
    """
    Procedure to infer _best_ Zslice  from linearly unmixed input.

    Parameters
    ------------
    raw_img:
        a ch,z,x,y - image containing florescent signal

    nuc_ch:
        channel with nuclei signal

    ch_to_agg:
        tuple of channels to aggregate for selecting Z

    Returns:
    -------------
    opt_z:
        the "0ptimal" z-slice which has the most signal intensity for downstream 2D segmentation
    """

    # median filter in 2D / convert to float 0-1.   get rid of the "residual"

    struct_img = min_max_intensity_normalization(raw_img[nuc_ch].copy())
    med_filter_size = 4
    nuclei = median_filter_slice_by_slice(struct_img, size=med_filter_size)

    gaussian_smoothing_sigma = 1.34
    gaussian_smoothing_truncate_range = 3.0
    nuclei = image_smoothing_gaussian_slice_by_slice(
        nuclei, sigma=gaussian_smoothing_sigma, truncate_range=gaussian_smoothing_truncate_range
    )
    thresh_factor = 0.9  # from cellProfiler
    thresh_min = 0.1
    thresh_max = 1.0
    struct_obj = apply_log_li_threshold(
        nuclei, thresh_factor=thresh_factor, thresh_min=thresh_min, thresh_max=thresh_max
    )

    optimal_Z = choose_agg_signal_zmax(raw_img, ch_to_agg, mask=struct_obj)
    print(f"choosing _optimal_ z-slice::: {optimal_Z}")
    return optimal_Z



import numpy as np
from typing import Dict
from pathlib import Path
import time

from infer_subc.constants import ER_CH
from infer_subc.core.file_io import export_inferred_organelle, import_inferred_organelle

from infer_subc.core.img import (
    select_channel_from_raw,
    scale_and_smooth,
    label_bool_as_uint16,
    fill_and_filter_linear_size,
    masked_object_thresh,
    filament_filter_3
)


##########################
#  infer_ER
##########################
def infer_ER(in_img: np.ndarray,
             ER_ch: int,
             median_sz: int,
             gauss_sig: float,
             MO_thresh_method: str,
             MO_cutoff_size: float,
             MO_thresh_adj: float,
             fil_scale_1: float,
             fil_cut_1: float,
             fil_scale_2: float, 
             fil_cut_2: float, 
             fil_scale_3: float, 
             fil_cut_3: float,
             fil_method: str,
             min_hole_w: int,
             max_hole_w: int,
             small_obj_w: int,
             fill_filter_method: str
             ) -> np.ndarray:
    """
    Procedure to infer ER from linearly unmixed input.

    Parameters
    ------------
    in_img: 
        a 3d image containing all the channels (CZYX)
    ER_ch:
        index of the ER channel in the input image
    median_sz: 
        width of median filter for signal
    gauss_sig: 
        sigma for gaussian smoothing of  signal
    mo_thresh_method: 
         which method to use for calculating global threshold. Options include:
         "triangle" (or "tri"), "median" (or "med"), and "ave_tri_med" (or "ave").
         "ave" refers the average of "triangle" threshold and "mean" threshold.
    mo_cutoff_size: 
        Masked Object threshold `size_min`
    mo_thresh_adjust: 
        Masked Object threshold `local_adjust`
    fil_scale_1: 
        scale (log_sigma) for filament filter
    fil_cutoff_1: 
        threshold for filament fitered threshold, associated to fil_scale_1
    fil_scale_2: 
        scale (log_sigma) for filament filter
    fil_cutoff_2: 
        threshold for filament fitered threshold, associated to fil_scale_2
    fil_scale_3: 
        scale (log_sigma) for filament filter
    fil_cutoff_3: 
        threshold for filament fitered threshold, associated to fil_scale_3
    fil_method:
        decision to process the filaments "slice-by-slice" or in "3D"
    min_hole_w: 
        minimum size for hole filling for cellmask signal post-processing
    max_hole_w: 
        hole filling cutoff for ER signal post-processing
    small_obj_w: 
        minimum object size cutoff for ER signal post-processing
    fill_filter_method:
        determines if small hole filling and small object removal should be run 'sice-by-slice' or in '3D'
    Returns
    -------------
    ER_object
        mask defined extent of ER object
    """

    ###################
    # EXTRACT
    ###################    
    ER = select_channel_from_raw(in_img, ER_ch)

    ###################
    # PRE_PROCESSING
    ###################    
    # er = normalized_edge_preserving_smoothing(er)
    struct_img =  scale_and_smooth(ER,
                                   median_size = median_sz, 
                                   gauss_sigma = gauss_sig)

    ###################
    # CORE_PROCESSING
    ###################
    bw1 = masked_object_thresh(struct_img, 
                                    global_method=MO_thresh_method, 
                                    cutoff_size=MO_cutoff_size, 
                                    local_adjust=MO_thresh_adj)

    bw2 = filament_filter_3(struct_img, fil_scale_1, fil_cut_1, fil_scale_2, fil_cut_2, fil_scale_3, fil_cut_3, fil_method)

    struct_obj = np.logical_or(bw1, bw2)
    
    ###################
    # POST_PROCESSING
    ################### 
    struct_obj = fill_and_filter_linear_size(struct_obj, 
                                             hole_min=min_hole_w, 
                                             hole_max=max_hole_w, 
                                             min_size=small_obj_w,
                                             method=fill_filter_method)

    ###################
    # LABELING
    ###################
    
    # ENSURE THAT there is ONLY ONE ER
    struct_obj = label_bool_as_uint16(struct_obj)

    return struct_obj 


##########################
#  fixed_infer_ER
##########################
def fixed_infer_ER(in_img: np.ndarray ) -> np.ndarray:
    """
    Procedure to infer endoplasmic rediculum from linearly unmixed input with *fixed parameters*

    Parameters
    ------------
    in_img: 
        a 3d image containing all the channels

    Returns
    -------------
    ER_object
        mask defined extent of ER object
    """
    ER_ch = 5
    median_sz = 3
    gauss_sig = 2.0
    MO_thresh_method = 'tri'
    MO_cutoff_size = 1200
    MO_thresh_adj = 0.7
    fil_scale_1 = 1
    fil_cut_1 = 0.005
    fil_scale_2 = 2
    fil_cut_2 = 0.005
    fil_scale_3 = 0
    fil_cut_3 = 0
    fil_method = "3D"
    min_hole_w = 0
    max_hole_w = 0
    small_obj_w = 4
    method = "3D"

    return infer_ER(
        in_img,
        ER_ch,
        median_sz,
        gauss_sig,
        MO_thresh_method,
        MO_cutoff_size,
        MO_thresh_adj,
        fil_scale_1,
        fil_cut_1,
        fil_scale_2,
        fil_cut_2,
        fil_scale_3,
        fil_cut_3,
        fil_method,
        min_hole_w,
        max_hole_w,
        small_obj_w,
        method)


def infer_and_export_ER(in_img: np.ndarray, meta_dict: Dict, out_data_path: Path) -> np.ndarray:
    """
    infer ER and write inferred ER to ome.tif file

    Parameters
    ------------
    in_img:
        a 3d  np.ndarray image of the inferred organelle (labels or boolean)
    meta_dict:
        dictionary of meta-data (ome)
    out_data_path:
        Path object where tiffs are written to

    Returns
    -------------
    exported file name

    """
    er = fixed_infer_ER(in_img)
    out_file_n = export_inferred_organelle(er, "ER", meta_dict, out_data_path)
    print(f"inferred ER. wrote {out_file_n}")
    return er


def get_ER(in_img: np.ndarray, meta_dict: Dict, out_data_path: Path) -> np.ndarray:
    """
    load endoplasmic_reticulum if it exists, otherwise calculate and write to ome.tif file

    Parameters
    ------------
    in_img:
        a 3d  np.ndarray image of the inferred organelle (labels or boolean)
    meta_dict:
        dictionary of meta-data (ome)
    out_data_path:
        Path object where tiffs are written to

    Returns
    -------------
    exported file name

    """

    try:
        er = import_inferred_organelle("ER", meta_dict, out_data_path)
    except:
        start = time.time()
        print("starting segmentation...")
        er = infer_and_export_ER(in_img, meta_dict, out_data_path)
        end = time.time()
        print(f"inferred (and exported) ER in ({(end - start):0.2f}) sec")

    return er



# centrosome routines
import numpy as np

from scipy.ndimage import distance_transform_edt, sum, minimum_filter, maximum_filter


def size_similarly(labels, secondary):
    """Size the secondary matrix similarly to the labels matrix

    labels - labels matrix
    secondary - a secondary image or labels matrix which might be of
                different size.
    Return the resized secondary matrix and a mask indicating what portion
    of the secondary matrix is bogus (manufactured values).

    Either the mask is all ones or the result is a copy, so you can
    modify the output within the unmasked region w/o destroying the original.
    """
    if labels.shape[:2] == secondary.shape[:2]:
        return secondary, np.ones(secondary.shape, bool)
    if labels.shape[0] <= secondary.shape[0] and labels.shape[1] <= secondary.shape[1]:
        if secondary.ndim == 2:
            return (
                secondary[: labels.shape[0], : labels.shape[1]],
                np.ones(labels.shape, bool),
            )
        else:
            return (
                secondary[: labels.shape[0], : labels.shape[1], :],
                np.ones(labels.shape, bool),
            )

    #
    # Some portion of the secondary matrix does not cover the labels
    #
    result = np.zeros(list(labels.shape) + list(secondary.shape[2:]), secondary.dtype)
    i_max = min(secondary.shape[0], labels.shape[0])
    j_max = min(secondary.shape[1], labels.shape[1])
    if secondary.ndim == 2:
        result[:i_max, :j_max] = secondary[:i_max, :j_max]
    else:
        result[:i_max, :j_max, :] = secondary[:i_max, :j_max, :]
    mask = np.zeros(labels.shape, bool)
    mask[:i_max, :j_max] = 1
    return result, mask


def distance_to_edge(labels):
    """Compute the distance of a pixel to the edge of its object

    labels - a labels matrix

    returns a matrix of distances
    """
    colors = color_labels(labels)
    max_color = np.max(colors)
    result = np.zeros(labels.shape)
    if max_color == 0:
        return result

    for i in range(1, max_color + 1):
        mask = colors == i
        result[mask] = distance_transform_edt(mask)[mask]
    return result


def color_labels(labels, distance_transform=False):
    """Color a labels matrix so that no adjacent labels have the same color

    distance_transform - if true, distance transform the labels to find out
         which objects are closest to each other.

    Create a label coloring matrix which assigns a color (1-n) to each pixel
    in the labels matrix such that all pixels similarly labeled are similarly
    colored and so that no similiarly colored, 8-connected pixels have
    different labels.

    You can use this function to partition the labels matrix into groups
    of objects that are not touching; you can then operate on masks
    and be assured that the pixels from one object won't interfere with
    pixels in another.

    returns the color matrix
    """
    if distance_transform:
        i, j = distance_transform_edt(labels == 0, return_distances=False, return_indices=True)
        dt_labels = labels[i, j]
    else:
        dt_labels = labels
    # Get the neighbors for each object
    v_count, v_index, v_neighbor = find_neighbors(dt_labels)
    # Quickly get rid of labels with no neighbors. Greedily assign
    # all of these a color of 1
    v_color = np.zeros(len(v_count) + 1, int)  # the color per object - zero is uncolored
    zero_count = v_count == 0
    if np.all(zero_count):
        # can assign all objects the same color
        return (labels != 0).astype(int)
    v_color[1:][zero_count] = 1
    v_count = v_count[~zero_count]
    v_index = v_index[~zero_count]
    v_label = np.argwhere(~zero_count).transpose()[0] + 1
    # If you process the most connected labels first and use a greedy
    # algorithm to preferentially assign a label to an existing color,
    # you'll get a coloring that uses 1+max(connections) at most.
    #
    # Welsh, "An upper bound for the chromatic number of a graph and
    # its application to timetabling problems", The Computer Journal, 10(1)
    # p 85 (1967)
    #
    sort_order = np.lexsort([-v_count])
    v_count = v_count[sort_order]
    v_index = v_index[sort_order]
    v_label = v_label[sort_order]
    for i in range(len(v_count)):
        neighbors = v_neighbor[v_index[i] : v_index[i] + v_count[i]]
        colors = np.unique(v_color[neighbors])
        if colors[0] == 0:
            if len(colors) == 1:
                # only one color and it's zero. All neighbors are unlabeled
                v_color[v_label[i]] = 1
                continue
            else:
                colors = colors[1:]
        # The colors of neighbors will be ordered, so there are two cases:
        # * all colors up to X appear - colors == np.arange(1,len(colors)+1)
        # * some color is missing - the color after the first missing will
        #   be mislabeled: colors[i] != np.arange(1, len(colors)+1)
        crange = np.arange(1, len(colors) + 1)
        misses = crange[colors != crange]
        if len(misses):
            color = misses[0]
        else:
            color = len(colors) + 1
        v_color[v_label[i]] = color
    return v_color[labels]


def find_neighbors(labels):
    """Find the set of objects that touch each object in a labels matrix

    Construct a "list", per-object, of the objects 8-connected adjacent
    to that object.
    Returns three 1-d arrays:
    * array of #'s of neighbors per object
    * array of indexes per object to that object's list of neighbors
    * array holding the neighbors.

    For instance, say 1 touches 2 and 3 and nobody touches 4. The arrays are:
    [ 2, 1, 1, 0], [ 0, 2, 3, 4], [ 2, 3, 1, 1]
    """
    max_label = np.max(labels)
    # Make a labels matrix with zeros around the edges so we can do index
    # offsets without worrying.
    #
    new_labels = np.zeros(np.array(labels.shape) + 2, labels.dtype)
    new_labels[1:-1, 1:-1] = labels
    labels = new_labels
    # Only consider the points that are next to others
    adjacent_mask = adjacent(labels)
    adjacent_i, adjacent_j = np.argwhere(adjacent_mask).transpose()
    # Get matching vectors of labels and neighbor labels for the 8
    # compass directions.
    count = len(adjacent_i)
    if count == 0:
        return (np.zeros(max_label, int), np.zeros(max_label, int), np.zeros(0, int))
    # The following bizarre construct does the following:
    # labels[adjacent_i, adjacent_j] looks up the label for each pixel
    # [...]*8 creates a list of 8 references to it
    # np.hstack concatenates, giving 8 repeats of the list
    v_label = np.hstack([labels[adjacent_i, adjacent_j]] * 8)
    v_neighbor = np.zeros(count * 8, int)
    index = 0
    for i, j in ((-1, -1), (-1, 0), (-1, 1), (0, -1), (0, 1), (1, -1), (1, 0), (1, 1)):
        v_neighbor[index : index + count] = labels[adjacent_i + i, adjacent_j + j]
        index += count
    #
    # sort by label and neighbor
    #
    sort_order = np.lexsort((v_neighbor, v_label))
    v_label = v_label[sort_order]
    v_neighbor = v_neighbor[sort_order]
    #
    # eliminate duplicates by comparing each element after the first one
    # to its previous
    #
    first_occurrence = np.ones(len(v_label), bool)
    first_occurrence[1:] = (v_label[1:] != v_label[:-1]) | (v_neighbor[1:] != v_neighbor[:-1])
    v_label = v_label[first_occurrence]
    v_neighbor = v_neighbor[first_occurrence]
    #
    # eliminate neighbor = self and neighbor = background
    #
    to_remove = (v_label == v_neighbor) | (v_neighbor == 0)
    v_label = v_label[~to_remove]
    v_neighbor = v_neighbor[~to_remove]
    #
    # The count of # of neighbors
    #
    v_count = fixup_scipy_ndimage_result(sum(np.ones(v_label.shape), v_label, np.arange(max_label, dtype=np.int32) + 1))
    v_count = v_count.astype(int)
    #
    # The index into v_neighbor
    #
    v_index = np.cumsum(v_count)
    v_index[1:] = v_index[:-1]
    v_index[0] = 0
    return (v_count, v_index, v_neighbor)


def fixup_scipy_ndimage_result(whatever_it_returned):
    """Convert a result from scipy.ndimage to a numpy array

    scipy.ndimage has the annoying habit of returning a single, bare
    value instead of an array if the indexes passed in are of length 1.
    For instance:
    scipy.ndimage.maximum(image, labels, [1]) returns a float
    but
    scipy.ndimage.maximum(image, labels, [1,2]) returns a list
    """
    if getattr(whatever_it_returned, "__getitem__", False):
        return np.array(whatever_it_returned)
    else:
        return np.array([whatever_it_returned])


def adjacent(labels):
    """Return a binary mask of all pixels which are adjacent to a pixel of
    a different label.

    """
    high = labels.max() + 1
    if high > np.iinfo(labels.dtype).max:
        labels = labels.astype(np.int32)
    image_with_high_background = labels.copy()
    image_with_high_background[labels == 0] = high
    min_label = minimum_filter(
        image_with_high_background,
        footprint=np.ones((3, 3), bool),
        mode="constant",
        cval=high,
    )
    max_label = maximum_filter(labels, footprint=np.ones((3, 3), bool), mode="constant", cval=0)
    return (min_label != max_label) & (labels > 0)




NAME = "infer_subc"

NUC_CH = 0
LYSO_CH = 1
MITO_CH = 2
GOLGI_CH = 3
PEROX_CH = 4
ER_CH = 5
LD_CH = 6
RESIDUAL_CH = 7

TEST_IMG_N = 5

ALL_CHANNELS = [NUC_CH, LYSO_CH, MITO_CH, GOLGI_CH, PEROX_CH, ER_CH, LD_CH, RESIDUAL_CH]

organelle_to_colname = {"nuc":"NU", "lyso": "LS", "mito":"MT", "golgi":"GL", "perox":"PR", "ER":"ER", "LD":"LD", "cell":"CM", "cyto":"CY", "nucleus": "N1","nuclei":"NU",}



import json

from pathlib import Path
from typing import Dict, List
from infer_subc.utils.directories import Directories
from infer_subc.workflow.segmenter_function import SegmenterFunction, FunctionParameter, WidgetType
from infer_subc.workflow.workflow_definition import WorkflowDefinition  # , PrebuiltWorkflowDefinition
from infer_subc.workflow.workflow_step import WorkflowStep, WorkflowStepCategory


class ConfigurationException(Exception):
    """
    Raised when errors are encountered reading from Configuration files
    """

    pass


class WorkflowConfig:
    """
    Provides access to structure workflow configuration
    """

    def __init__(self):
        self._all_functions = None
        self._available_workflow_names = None

    def get_available_workflows(self) -> List[str]:
        """
        Get the list of all workflows available through configuration
        """
        if self._available_workflow_names is None:
            json_list = sorted(Directories.get_structure_config_dir().glob("conf_*.json"))
            self._available_workflow_names = [p.stem[5:] for p in json_list]

        return self._available_workflow_names

    def get_all_functions(self) -> List[SegmenterFunction]:
        """
        Get the list of all available Functions from configuration
        """
        if self._all_functions is None:
            path = Directories.get_structure_config_dir() / "all_functions.json"

            try:
                with open(path) as file:
                    obj = json.load(file)
                    self._all_functions = self._all_functions_decoder(obj)
            except Exception as ex:
                raise ConfigurationException(f"Error reading json configuration from {path}") from ex

        return self._all_functions

    def get_workflow_definition(self, workflow_name: str) -> WorkflowDefinition:  # PrebuiltWorkflowDefinition:
        """
        Get a WorkflowDefinition for the given workflow from the corresponding
        prebuilt json structure config
        """
        if workflow_name is None or len(workflow_name.strip()) == 0:
            raise ValueError("workflow_name cannot be empty")

        if workflow_name not in self.get_available_workflows():
            raise ValueError(f"No workflow configuration available for {workflow_name}")

        path = Directories.get_structure_config_dir() / f"conf_{workflow_name}.json"

        return self.get_workflow_definition_from_config_file(path, workflow_name, prebuilt=True)

    def get_workflow_definition_from_config_file(
        self, file_path: Path, workflow_name: str = None, prebuilt: bool = False
    ) -> WorkflowDefinition:
        """
        Get a WorkflowDefinition based off the given json configuration file
        """
        if file_path.suffix.lower() != ".json":
            raise ValueError("Workflow configuration file must be a json file with .json file extension.")

        with open(file_path) as file:
            try:
                obj = json.load(file)
                # print(obj)
                return self._workflow_decoder(obj, workflow_name or file_path.stem, prebuilt)
            except Exception as ex:
                raise ConfigurationException(f"Error reading json configuration from {file_path}") from ex

    def save_workflow_definition_as_json(self, workflow_definition: WorkflowDefinition, output_file_path: Path):
        """
        Save a WorkflowDefinition as a json config file
        """
        if output_file_path.suffix.lower() != ".json":
            raise ValueError("Workflow configuration file save path must have a .json extension.")

        with open(output_file_path, "w") as file:
            json.dump(self._workflow_encoder(workflow_definition), file, indent=4, sort_keys=True)

    def _all_functions_decoder(self, obj: Dict) -> List[SegmenterFunction]:
        """
        Decode Functions config (all_functions.json)
        """

        def build_function_parameter(name: str, data: Dict):
            return FunctionParameter(
                name=name,
                widget_type=WidgetType.from_str(data["widget_type"]),
                data_type=data["data_type"],
                min_value=data.get("min", None),
                max_value=data.get("max", None),
                increment=data.get("increment", None),
                options=data.get("options", None),
            )

        functions = list()
        for function_k, function_v in obj.items():
            function = SegmenterFunction(
                name=function_k,
                display_name=function_v["name"],
                function=function_v["python::function"],
                module=function_v["python::module"],
            )

            if function_v.get("parameters") is not None and len(function_v["parameters"]) > 0:
                params = dict()

                for param_k, param_v in function_v["parameters"].items():
                    param_name = param_k
                    params[param_name] = list()

                    if isinstance(param_v, dict):
                        params[param_name].append(build_function_parameter(param_name, param_v))
                    elif isinstance(param_v, list):
                        for item in param_v:
                            params[param_name].append(build_function_parameter(param_name, item))

                function.parameters = params

            functions.append(function)

        return functions

    def _workflow_decoder(self, obj: Dict, workflow_name: str, prebuilt: bool = False) -> WorkflowDefinition:
        """
        Decode Workflow config (conf_{workflow_name}.json)
        """
        functions = self.get_all_functions()
        steps: List[WorkflowStep] = list()
        for step_k, step_v in obj.items():
            # print(f"step_k{step_k}  - step_v{step_v}")

            step_number = int(step_k)
            function_id = step_v["function"]
            function = next(filter(lambda f: f.name == function_id, functions), None)

            if function is None:
                raise ConfigurationException(
                    f"Could not find a Segmenter function matching the function identifier <{function_id}>."
                )

            if isinstance(step_v["parent"], list):
                parent = step_v["parent"]
            else:
                parent = [step_v["parent"]]

            step = WorkflowStep(
                category=WorkflowStepCategory.from_str(step_v["category"]),
                function=function,
                step_number=step_number,
                parent=parent,
            )

            if step_v.get("parameter_values") is not None and len(step_v["parameter_values"]) > 0:
                param_defaults = dict()

                for param_k, param_v in step_v["parameter_values"].items():
                    param_name = param_k
                    param_defaults[param_name] = param_v

                step.parameter_values = param_defaults
            # print(f"adding step {step_number}")
            steps.append(step)

        steps.sort(key=lambda s: s.step_number)

        return WorkflowDefinition(workflow_name, steps, prebuilt=prebuilt)

    def _workflow_encoder(self, workflow_definition: WorkflowDefinition) -> Dict:
        """
        Encode a WorkflowDefinition to a json dictionary
        """

        # TODO add header / version ?
        result = dict()
        for step in workflow_definition.steps:
            step_number = str(step.step_number)
            parent = step.parent[0] if len(step.parent) == 1 else step.parent

            step_dict = {
                step_number: {"function": step.function.name, "category": step.category.value, "parent": parent}
            }
            if step.parameter_values is not None:
                step_dict[step_number].update({"parameter_values": step.parameter_values})

            result.update(step_dict)

        return result



from typing import Union, Optional, Dict, List

from pathlib import Path
import numpy as np

from infer_subc.core.file_io import export_inferred_organelle, read_czi_image, list_image_files, export_tiff, read_tiff_image


from infer_subc.constants import (
    TEST_IMG_N,
    NUC_CH,
    LYSO_CH,
    MITO_CH,
    GOLGI_CH,
    PEROX_CH,
    ER_CH,
    LD_CH,
    RESIDUAL_CH,
)


from infer_subc.core.img import label_uint16





from infer_subc.organelles import (
    fixed_infer_cellmask_fromaggr,
    fixed_infer_nuclei,
    infer_cytoplasm,
    fixed_infer_lyso,
    fixed_infer_mito,
    fixed_infer_golgi,
    fixed_infer_ER,
    fixed_infer_perox,
    fixed_infer_LD,
)


def explode_mask(mask_path: Union[Path,str], postfix: str= "masks", im_type: str = ".tiff") -> bool:
    """ 
    Split a 3 channel 'masks' file into 'nuc', 'cell', and 'cyto' images.  WARNING: requires the channels to be nuc = 0, cell = 1, and cyto = 2
    TODO: add logging instead of printing
        append tiffcomments with provenance
    """
    if isinstance(mask_path, str): mask_path = Path(mask_path)
    # load image 
    full_stem = mask_path.stem
    if full_stem.endswith(postfix):
        stem = full_stem.rstrip(postfix)
        image = read_tiff_image(mask_path)
        assert image.shape[0]==3
        
        # make into np.uint16 labels
        nuclei = label_uint16(image[0])
        # export as np.uint8 (255)
        cellmask = image[1]>0            
        cytoplasm = image[2]>0

        # write wasks
        root_stem = mask_path.parent / stem
        # ret1 = imwrite(f"{root}nuclei{stem}", nuclei)
        ret1 = export_tiff(nuclei, f"{stem}nuc", mask_path.parent, None)
        # ret2 = imwrite(f"{root}cellmask{stem}", cellmask)
        ret2 = export_tiff(cellmask, f"{stem}cell", mask_path.parent, None)
        # ret3 = imwrite(f"{root}cytosol{stem}", cytosol)
        ret3 = export_tiff(cytoplasm, f"{stem}cyto", mask_path.parent, None)

        # print(f"wrote {stem}-{{nuclei,cellmask,cyto}}")
        return True
    else:
        return False
    

def explode_masks(root_path: Union[Path,str], postfix: str= "masks", im_type: str = ".tiff"):
    """  
    TODO: add loggin instead of printing
        append tiffcomments with provenance
    """
    if isinstance(root_path, str): root_path = Path(root_path)

    img_file_list = list_image_files(root_path,im_type, postfix)
    wrote_cnt = 0
    for img_f in img_file_list:
        if explode_mask(img_f, postfix=postfix, im_type=im_type): 
            wrote_cnt += 1
        else: 
            print(f"failed to explode {img_f}")
    # else:
    #     print(f"how thefark!!! {img_f}")
   

    print(f"exploded {wrote_cnt*100./len(img_file_list)} pct of {len(img_file_list)} files")
    return wrote_cnt



def find_segmentation_tiff_files(prototype:Union[Path,str], organelles: List[str], int_path: Union[Path,str]) -> Dict:
    """
    find the nescessary image files based on protype, the organelles involved, and paths
    """

    # raw
    prototype = Path(prototype)
    if not prototype.exists():
        print(f"bad prototype. please choose an existing `raw` file as prototype")
        return dict()
    # make sure protoype ends with czi

    out_files = {"raw":prototype}

    int_path = Path(int_path) 
    # raw
    if not int_path.is_dir():
        print(f"bad path argument. please choose an existing path containing organelle segmentations")
        return out_files
    
    # cyto, cellmask
    cyto_nm = int_path / f"{prototype.stem}-cyto.tiff"
    if cyto_nm.exists():
        out_files["cyto"] = cyto_nm
    else:
        print(f"cytosol mask not found.  We'll try to extract from masks ")
        if explode_mask(int_path / f"{prototype.stem}-masks.tiff"): 
            out_files["cyto"] = cyto_nm
        else: 
            print(f"failed to explode {prototype.stem}-masks.tiff")
            return out_files
    
    cellmask_nm = int_path / f"{prototype.stem}-cell.tiff"
    if  cellmask_nm.exists():
        out_files["cell"] = cellmask_nm
    else:
        print(f"cellmask file not found in {int_path} returning")
        out_files["cell"] = None

    # organelles
    for org_n in organelles:
        org_name = Path(int_path) / f"{prototype.stem}-{org_n}.tiff"
        if org_name.exists(): 
            out_files[org_n] = org_name
        else: 
            print(f"{org_n} .tiff file not found in {int_path} returning")
            out_files[org_n] = None
    
    if "nuc" not in organelles:
        nuc_nm = int_path / f"{prototype.stem}-nuc.tiff"
        if  nuc_nm.exists():
            out_files["nuc"] = nuc_nm
        else:
            print(f"nuc file not found in {int_path} returning")
            out_files["nuc"] = None



    return out_files


###########
# infer organelles
##########
def fixed_infer_organelles(img_data):
    """
    wrapper to infer all organelles from a single multi-channel image
    """
    # ch_to_agg = (LYSO_CH, MITO_CH, GOLGI_CH, PEROX_CH, ER_CH, LD_CH)

    # nuc_ch = NUC_CH
    # optimal_Z = find_optimal_Z(img_data, nuc_ch, ch_to_agg)
    # optimal_Z = fixed_find_optimal_Z(img_data)
    # # Stage 1:  nuclei, cellmask, cytoplasm
    # img_2D = fixed_get_optimal_Z_image(img_data)

    cellmask = fixed_infer_cellmask_fromaggr(img_data)

    nuclei_object = fixed_infer_nuclei(img_data, cellmask)

    cytoplasm_mask = infer_cytoplasm(nuclei_object, cellmask)

    # cyto masked objects.
    lyso_object = fixed_infer_lyso(img_data, cytoplasm_mask)
    mito_object = fixed_infer_mito(img_data, cytoplasm_mask)
    golgi_object = fixed_infer_golgi(img_data, cytoplasm_mask)
    peroxi_object = fixed_infer_perox(img_data, cytoplasm_mask)
    er_object = fixed_infer_ER(img_data, cytoplasm_mask)
    LD_object = fixed_infer_LD(img_data, cytoplasm_mask)

    img_layers = [
        nuclei_object,
        lyso_object,
        mito_object,
        golgi_object,
        peroxi_object,
        er_object,
        LD_object,
        cellmask,
        cytoplasm_mask,
    ]

    layer_names = [
        "nuclei",
        "lyso",
        "mitochondria",
        "golgi",
        "peroxisome",
        "er",
        "LD_body",
        "cell",
        "cytoplasm_mask",
    ]
    # TODO: pack outputs into something napari readable
    img_out = np.stack(img_layers, axis=0)
    return (img_out, layer_names)


def batch_process_all_czi(data_root_path, source_dir: Union[Path, str] = "raw"):
    """ """
    # linearly unmixed ".czi" files are here
    data_path = data_root_path / "raw"
    im_type = ".czi"
    # get the list of all files
    img_file_list = list_image_files(data_path, im_type)
    files_generated = []
    for czi_file in img_file_list:
        out_fn = process_czi_image(czi_file)
        files_generated.append(out_fn)

    print(f"generated {len(files_generated)} ")
    return files_generated


def process_czi_image(czi_file_name, data_root_path):
    """wrapper for processing"""

    img_data, meta_dict = read_czi_image(czi_file_name)
    # # get some top-level info about the RAW data
    # channel_names = meta_dict['name']
    # img = meta_dict['metadata']['aicsimage']
    # scale = meta_dict['scale']
    # channel_axis = meta_dict['channel_axis']

    inferred_organelles, layer_names, optimal_Z = fixed_infer_organelles(img_data)
    out_file_n = export_inferred_organelle(inferred_organelles, layer_names, meta_dict, data_root_path)

    ## TODO:  collect stats...

    return out_file_n


def stack_organelle_objects(
    cellmask,
    nuclei_object,
    cytoplasm_mask,
    lyso_object,
    mito_object,
    golgi_object,
    peroxi_object,
    er_object,
    LD_object,
) -> np.ndarray:
    """wrapper to stack the inferred objects into a single numpy.ndimage"""
    img_layers = [
        cellmask,
        nuclei_object,
        cytoplasm_mask,
        lyso_object,
        mito_object,
        golgi_object,
        peroxi_object,
        er_object,
        LD_object,
    ]
    return np.stack(img_layers, axis=0)


def stack_organelle_layers(*layers) -> np.ndarray:
    """wrapper to stack the inferred objects into a single numpy.ndimage"""

    return np.stack(layers, axis=0)



import numpy as np
from typing import Dict, Optional
from pathlib import Path
import time

from aicssegmentation.core.seg_dot import dot_3d_wrapper, dot_2d_slice_by_slice_wrapper
from aicssegmentation.core.utils import topology_preserving_thinning

from infer_subc.constants import GOLGI_CH
from infer_subc.core.file_io import export_inferred_organelle, import_inferred_organelle
from infer_subc.core.img import (
    fill_and_filter_linear_size,
    select_channel_from_raw,
    masked_object_thresh,
    scale_and_smooth,
    label_uint16,
    dot_filter_3
)


##########################
#  infer_golgi
##########################
def infer_golgi(
            in_img: np.ndarray,
            golgi_ch: int,
            median_sz: int,
            gauss_sig: float,
            mo_method: str,
            mo_adjust: float,
            mo_cutoff_size: int,
            min_thickness: int,
            thin_dist: int,
            dot_scale_1: float,
            dot_cut_1: float,
            dot_scale_2: float,
            dot_cut_2: float,
            dot_scale_3: float,
            dot_cut_3: float,
            dot_method: str,
            min_hole_w: int,
            max_hole_w: int,
            small_obj_w: int,
            fill_filter_method: str
        ) -> np.ndarray:

    """
    Procedure to infer golgi from linearly unmixed input.

   Parameters
    ------------
    in_img: 
        a 3d image containing all the channels
    median_sz: 
        width of median filter for signal
    mo_method: 
         which method to use for calculating global threshold. Options include:
         "triangle" (or "tri"), "median" (or "med"), and "ave_tri_med" (or "ave").
         "ave" refers the average of "triangle" threshold and "mean" threshold.
    mo_adjust: 
        Masked Object threshold `local_adjust`
    mo_cutoff_size: 
        Masked Object threshold `size_min`
    min_thinkness: 
        Half of the minimum width you want to keep from being thinned.
        For example, when the object width is smaller than 4, you don't
        want to make this part even thinner (may break the thin object
        and alter the topology), you can set this value as 2.
    thin_dist: 
        the amount to thin (has to be an positive integer). The number of
         pixels to be removed from outter boundary towards center.
    dot_scale: 
        scales (log_sigma) for dot filter (1,2, and 3)
    dot_cut: 
        threshold for dot filter thresholds (1,2,and 3)
    small_obj_w: 
        minimu object size cutoff for nuclei post-processing
    
    Returns
    -------------
    golgi_object
        mask defined extent of golgi object
    """

    ###################
    # EXTRACT
    ###################    
    golgi = select_channel_from_raw(in_img, golgi_ch)

    ###################
    # PRE_PROCESSING
    ###################    
    golgi =  scale_and_smooth(golgi,
                              median_size = median_sz, 
                              gauss_sigma = gauss_sig)
    ###################
    # CORE_PROCESSING
    ###################
    bw = masked_object_thresh(golgi, global_method=mo_method, cutoff_size=mo_cutoff_size, local_adjust=mo_adjust)

    bw_thin = topology_preserving_thinning(bw, min_thickness, thin_dist)

    bw_extra = dot_filter_3(golgi, dot_scale_1, dot_cut_1, dot_scale_2, dot_cut_2, dot_scale_3, dot_cut_3, dot_method)

    bw = np.logical_or(bw_extra, bw_thin)
    ###################
    # POST_PROCESSING
    ###################
    struct_obj = fill_and_filter_linear_size(bw, 
                                             hole_min=min_hole_w, 
                                             hole_max=max_hole_w, 
                                             min_size=small_obj_w,
                                             method=fill_filter_method)

    ###################
    # LABELING
    ###################
    struct_obj1 = label_uint16(struct_obj)

    return struct_obj1


# def infer_golgi(
#     in_img: np.ndarray,
#     median_sz: int,
#     gauss_sig: float,
#     mo_method: str,
#     mo_adjust: float,
#     mo_cutoff_size: int,
#     min_thickness: int,
#     thin: int,
#     dot_scale: float,
#     dot_cut: float,
#     small_obj_w: int,
# ) -> np.ndarray:
#     """
#      Procedure to infer golgi from linearly unmixed input.

#     Parameters
#      ------------
#      in_img:
#          a 3d image containing all the channels
#      median_sz:
#          width of median filter for signal
#      mo_method:
#           which method to use for calculating global threshold. Options include:
#           "triangle" (or "tri"), "median" (or "med"), and "ave_tri_med" (or "ave").
#           "ave" refers the average of "triangle" threshold and "mean" threshold.
#      mo_adjust:
#          Masked Object threshold `local_adjust`
#      mo_cutoff_size:
#          Masked Object threshold `size_min`
#      min_thinkness:
#          Half of the minimum width you want to keep from being thinned.
#          For example, when the object width is smaller than 4, you don't
#          want to make this part even thinner (may break the thin object
#          and alter the topology), you can set this value as 2.
#      thin:
#          the amount to thin (has to be an positive integer). The number of
#           pixels to be removed from outter boundary towards center.
#      dot_scale:
#          scales (log_sigma) for dot filter (1,2, and 3)
#      dot_cut:
#          threshold for dot filter thresholds (1,2,and 3)
#      small_obj_w:
#          minimu object size cutoff for nuclei post-processing

#      Returns
#      -------------
#      golgi_object
#          mask defined extent of golgi object
#     """
#     golgi_ch = GOLGI_CH

#     ###################
#     # EXTRACT
#     ###################
#     golgi = select_channel_from_raw(in_img, golgi_ch)

#     ###################
#     # PRE_PROCESSING
#     ###################
#     golgi = scale_and_smooth(golgi, median_sz=median_sz, gauss_sig=gauss_sig)
#     ###################
#     # CORE_PROCESSING
#     ###################
#     # bw = MO(golgi, global_thresh_method=thresh_method, object_minArea=obj_min_area)
#     bw = masked_object_thresh(golgi, th_method=mo_method, cutoff_size=mo_cutoff_size, th_adjust=mo_adjust)

#     bw_thin = topology_preserving_thinning(bw, min_thickness, thin)

#     s3_param = [(dot_cut, dot_scale)]
#     bw_extra = dot_2d_slice_by_slice_wrapper(golgi, s3_param)
#     # bw_extra = dot_3d_wrapper(golgi, s3_param)

#     bw = np.logical_or(bw_extra, bw_thin)
#     ###################
#     # POST_PROCESSING
#     ###################
#     struct_obj = size_filter_linear_size(bw, min_size=small_obj_w, connectivity=1)

#     return label_uint16(struct_obj)


##########################
#  fixed_infer_golgi
##########################
def fixed_infer_golgi(in_img: np.ndarray ) -> np.ndarray:
    """
    Procedure to infer golgi from linearly unmixed input.

    Parameters
    ------------
    in_img: 
        a 3d image containing all the channels
    Returns
    -------------
   golgi_object
        mask defined extent of golgi object
    """

    golgi_ch = 3
    median_sz = 4
    gauss_sig = 1.34
    mo_method = 'tri'
    mo_adjust = 1
    mo_cutoff_size = 1200 
    min_thickness = 1.6
    thin_dist = 1
    dot_scale_1 = 1.6
    dot_cut_1 = 0.02
    dot_scale_2 = 0
    dot_cut_2 = 0
    dot_scale_3 = 0
    dot_cut_3 = 0
    dot_method = '3D'
    min_hole_w = 0
    max_hole_w = 0
    small_obj_w = 3
    fill_filter_method = "3D"

    return infer_golgi(
        in_img,
        golgi_ch,
        median_sz,
        gauss_sig,
        mo_method,
        mo_adjust,
        mo_cutoff_size,
        min_thickness,
        thin_dist,
        dot_scale_1,
        dot_cut_1,
        dot_scale_2,
        dot_cut_2,
        dot_scale_3,
        dot_cut_3,
        dot_method,
        min_hole_w,
        max_hole_w,
        small_obj_w,
        fill_filter_method)


# def fixed_infer_golgi(in_img: np.ndarray, cytoplasm_mask: Optional[np.ndarray] = None) -> np.ndarray:
#     """
#      Procedure to infer golgi from linearly unmixed input.

#      Parameters
#      ------------
#      in_img:
#          a 3d image containing all the channels
#      Returns
#      -------------
#     golgi_object
#          mask defined extent of golgi object
#     """

#     median_sz = 4
#     gauss_sig = 1.34
#     mo_method = "tri"
#     mo_adjust = 0.90
#     mo_cutoff_size = 1200
#     min_thickness = 1.6
#     thin = 1
#     dot_scale = 1.6
#     dot_cut = 0.02
#     small_obj_w = 3

#     return infer_golgi(
#         in_img,
#         median_sz,
#         gauss_sig,
#         mo_method,
#         mo_adjust,
#         mo_cutoff_size,
#         min_thickness,
#         thin,
#         dot_scale,
#         dot_cut,
#         small_obj_w,
#     )


def infer_and_export_golgi(in_img: np.ndarray, meta_dict: Dict, out_data_path: Path) -> np.ndarray:
    """
    infer golgi and write inferred golgi to ome.tif file

    Parameters
    ------------
    in_img:
        a 3d  np.ndarray image of the inferred organelle (labels or boolean)
    meta_dict:
        dictionary of meta-data (ome)
    out_data_path:
        Path object where tiffs are written to

    Returns
    -------------
    exported file name

    """
    golgi = fixed_infer_golgi(in_img)
    out_file_n = export_inferred_organelle(golgi, "golgi", meta_dict, out_data_path)
    print(f"inferred golgi. wrote {out_file_n}")
    return golgi


def get_golgi(in_img: np.ndarray, meta_dict: Dict, out_data_path: Path) -> np.ndarray:
    """
    load golgi if it exists, otherwise calculate and write to ome.tif file

    Parameters
    ------------
    in_img:
        a 3d  np.ndarray image of the inferred organelle (labels or boolean)
    meta_dict:
        dictionary of meta-data (ome)
    out_data_path:
        Path object where tiffs are written to

    Returns
    -------------
    exported file name

    """

    try:
        golgi = import_inferred_organelle("golgi", meta_dict, out_data_path)
    except:
        start = time.time()
        print("starting segmentation...")
        golgi = infer_and_export_golgi(in_img, meta_dict, out_data_path)
        end = time.time()
        print(f"inferred (and exported) golgi in ({(end - start):0.2f}) sec")

    return golgi
